//AstroCalc.h
//
//Copyright (C) 2003 mikesag
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//This is the interface between the main program and the Swiss Ephemeris.
//To use another astrology library, edit this file and Translation.cpp/.h,
//It also simplifies the use of the Swiss Ephemeris considerably, by
//encapsulating all the temporary variables needed when calling.

#ifndef __ASTROCALC_H__
#define __ASTROCALC_H__

#include "AstroConst.h"

#define USE_TRUE 0
#define USE_MEAN 1
#define USE_APOG USE_TRUE

#define HOUSEARRSIZE  14

// Chart types, as used in the array chart[] and chartdata[]
#define NCHARTTYPES       8

#define CHART_NATAL1      0
#define CHART_NATAL2      1
#define CHART_TRANSITS    2
#define CHART_PROGRESSED1 3
#define CHART_PROGRESSED2 4
#define CHART_DAVISON     5
#define CHART_COMPOSITE   6
#define CHART_DEFAULT     7

#define SQ(n) ((n)*(n))
#define JD_DAYS_IN_YEAR 365.24219893

struct trSignPos
{
	int degrees;
	int sign;      // in internal format, *not* 1-12!
	int minutes;
	double seconds;
};

struct trHousePos
{
	int hsystem;   // so we know which house system this was computed under!
	double cusp;
	trSignPos sgn;
};

struct trPlanet
{
	double longitude;
	double latitude;
	double dist;
	double long_speed;
	double lat_speed;
	double dist_speed;
	trSignPos sgn;
	int hsystem; // house system
	int in_house;
};

// Putting it all together...
// By storing everything in a structure like this, we reduce the need to
// recompute astrological info over and over, in effect "caching' the info
struct trChartInfo
{
	double sidereal_time;           // Sidereal time; used for getting the ARMC
	double armc;                    // ARMC, used for composite and progressed charts
	trHousePos house[HOUSEARRSIZE]; // 0 is blank 1-12 are house cusps by degree, and 13 is
	                   // the same as 1, to make it easier to compute ranges
	trPlanet planet[NTOTOBJECTS];  // store possibly all planets at once
	                               // note that not every single object is there,
	                               // just those from SUN to SUN+NTOTOBJECTS-1
};

// Days, Minutes Seconds
struct trDMS
{
	union {
		int deg;
		int hours;  // 24-hour format only
	};
	int min;
	double sec;
};

// Hours, Minutes, Seconds
typedef trDMS trHMS;

struct trDate
{
	int month; // 1-based, not 0-based!
	int day;   // ditto
	int year;  // Y2K compliant, please
};

struct trTimeDatePlace
{
	trDMS latitude;  // Positive is Northern Hemisphere; Negative is Southern
	trDMS longitude; // Negative is Western Hemisphere; Positive is Eastern
	trHMS time;
	int timezone;    // Negative is Western Hemisphere; Positive is Eastern
	int is_dst;
	int is_progressed;
	trDate date;
};

void set_node_type(int n);
int  get_node_type();
void set_apog_type(int n);
int  get_apog_type();

// Utility function to convert from a raw degree to "xx sgn yy"-formatted sign
void split_longitute(double longitude, trSignPos *pos, int round_min);

// Utility function to merge degree, minute, and second values into a single
// double-precious degree value
double merge_dms(trDMS *dms);

// Converse of above
void split_dms(double x, trDMS *dms);

// Find the midpoints for composite and Davison charts, as well as any two
// geographic coordinates (DMS)
void midpoint_dms(trDMS *dms1, trDMS *dms2, trDMS *midpt);
void midpoint_tdp(trTimeDatePlace *tdp1, trTimeDatePlace *tdp2, trTimeDatePlace *midtdp);
void midpoint_planets(trPlanet *pl1, trPlanet *pl2, trChartInfo *midchart);

// Progress a chart
void progress_tdp(trTimeDatePlace *tdp_orig, trTimeDatePlace *tdp_now,
				  trTimeDatePlace *tdp_progressed);

// Negate the largest non-zero component of the DMS structure
void negate_dms(trDMS *dms);
int is_negative_dms(const trDMS *dms);

// Conversions between Julian date doubles and Gregorian date structures
double to_juldate(trTimeDatePlace *tdp);
void from_juldate(double juldate, int timezone, int is_dst, trTimeDatePlace *tdp);
double calc_juldate(int y, int mon, int d, int h, int min, double sec,
					int timezone, int is_dst);
void convert_juldate(double juldate, int timezone, int is_dst,
					 int *y, int *mon, int *d, int *h, int *min, double *sec);
void correct_for_dst(trTimeDatePlace *tdp);
void make_dst(trTimeDatePlace *tdp);

// Call this before calling any other function from this module!
void init_astro_library(char *argv0);

// Call this before exiting the program
void close_astro_library();

// Check if error occured on last call
bool ac_error_occured();

// Which planet/house/whatever did the error occur at?
int ac_find_error();

// If there was an error, return the error message
char *ac_error_msg();

// Main calculation routines; calculates everything you need to draw a chart!
void astro_calc(trTimeDatePlace *tdp, int hsys, trChartInfo *pinf);
void astro_calc_progressed(trTimeDatePlace *tdp1, trTimeDatePlace *tdp2,
						   trTimeDatePlace *tdppr, int hsys, trChartInfo *pinf);
void astro_calc_composite(trTimeDatePlace *tdp1, trTimeDatePlace *tdp2,
						  double armc1, double armc2, int hsys, trChartInfo *pinf);

// Smaller routines that calculate one thing at a time; for use when calculating
// something or displaying just one more piece of info
void get_planets_no_houses(trTimeDatePlace *tdp, trPlanet pl[NTOTOBJECTS]); // All but house
void get_houses(trTimeDatePlace *tdp, int hsys, trHousePos house[HOUSEARRSIZE],
				double *armc, double *sidereal_time);
void get_houses_ex(trTimeDatePlace *tdp, int hsys, trHousePos house[HOUSEARRSIZE],
				int by_armc, double *armc, double *sidereal_time, double latitude, double eps);
				// If we can calculate one house, we get the rest for free, so why not?
				// Note that armc is output iff !by_armc, and input iff by_armc,
				// and that latitude and eps are only used if by_armc is true
void get_planet(int npl, trTimeDatePlace *tdp, trPlanet *pl);  // Fills entire trPlanet structure
int  get_which_house(int npl, trChartInfo *pinf, trHousePos house[HOUSEARRSIZE]);
void fill_planets_in_houses(trChartInfo *pinf, int hsys); // Finds which planets are in which houses
double get_ecliptic(trTimeDatePlace *tdp);

// Finally, wrappers when you really, really only want one piece of info.  Note that
// this can get expensive if called repeatedly so please try to be efficient by
// calling one of the above functions instead.  They're marked with "ac_" so as not to
// be confused with the less wasteful functions above.  Note that a different astrology
// library might make the costs different, so nothing is set in stone.
double ac_get_planet_longitude(int npl, trTimeDatePlace *tdp);  // Longitude only (sign)
double ac_get_house_pos(int nhouse, trTimeDatePlace *tdp, int hsys);


#endif
