//TransitsChart.cpp: This is where the chart is drawed
//
//Copyright (C) 2003 mikesag
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "wx/wxprec.h"

#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

#include <math.h>

#include "wx/settings.h"

#include "AstroConst.h"
#include "AstroCalc.h"

#include "TurtleDraw.h"
#include "Translation.h"

#include "TransitsChart.h"

#define MINSPREAD 5.5

static const double N_PI = acos(0)*2.0;

// Custom wxWindows array initialization

#include "wx/arrimpl.cpp"
WX_DEFINE_OBJARRAY(aspect_configuration_array);

#ifndef inline
#define inline
#endif

inline static int iround(const double x)
{
	if(x > 0.0)
		return ((int)floor(x + 0.5));
	else
		return ((int)ceil(x - 0.5));
}

void aspect_configuration::Clear()
{
	aspect.Clear();
	planet.Clear();
	aspect.Shrink();
	planet.Shrink();
}

aspect_configuration::~aspect_configuration()
{
	Clear();
}

// ----------------------------------------------------------------------------
// chart window
// ----------------------------------------------------------------------------

IMPLEMENT_CLASS(CChartCanvas, wxWindow)
BEGIN_EVENT_TABLE(CChartCanvas, wxWindow)
	EVT_SIZE         (CChartCanvas::OnResize)
    EVT_PAINT        (CChartCanvas::OnPaint)
    EVT_MOTION       (CChartCanvas::OnMouseMove)
	EVT_LEFT_DOWN    (CChartCanvas::OnMouseDown)
	EVT_LEAVE_WINDOW (CChartCanvas::OnMouseLeaveWindow)
END_EVENT_TABLE()

CChartCanvas::CChartCanvas(wxFrame *parent)
        : wxWindow(parent, -1, wxDefaultPosition, wxSize(100,100))
{
	int i, j, k;
    owner = parent;
	SetBackgroundColour(*wxBLACK);
	for(i = 0; i < NCHARTTYPES; i++)
		for(j = 0; j < (NTOTOBJECTS + 12); j++)
			restricted[i][j] = FALSE;

	for(k = 0; k < NASPECTS; k++)
	{
		// These will both be overwritten by the parent frame,
		// so make it obvious if they're not
		aspect_orb[k] = 10.0;
		r_aspects[k] = FALSE;
	}

	n_aspect_cache = 0;
	chart1 = chart2 = 0;
	mode = NATAL_CHART;
	infopane = NULL;
}

void CChartCanvas::Initialize()
{
	if(infopane != NULL)
		infopane->SetAspects(n_aspect_cache, aspect_cache);
}

void CChartCanvas::SetMode(int newmode, int n, int n1, int n2,
						   bool recalc_aspects, bool do_refresh)
{
	int oldmode = mode;
	mode = newmode;
	if(IS_ONE_CHART(newmode))
		chart1 = n;
	else
		chart1 = n1;
	chart2 = n2;
	infopane->SetChartSel(n, n1, n2);
	if(oldmode != newmode)
		infopane->SetViewMode(newmode);
	if(IS_ONE_CHART(newmode))
		infopane->SetMode(infopane->GetMode(), infopane->GetObj(), chart1);

	// Not needed if SetChartData is called afterwards
	if(recalc_aspects)
	{
		if(IS_ONE_CHART(newmode))
			CalculateAspects(n, n, FALSE);
		else
			CalculateAspects(n1, n2, TRUE);

		if(newmode & ASPCONFIG_CHART)
		{
			wxASSERT(IS_ONE_CHART(newmode));
			CalculateAspectConfigurations(n);
		}
		else
		{
			aspect_conf.Clear();
		}
	}

	if(do_refresh)
	{
		Refresh(FALSE);
		infopane->Refresh(FALSE);
	}
}

void CChartCanvas::SetChartData(trChartInfo ci[NCHARTTYPES], trTimeDatePlace cd[NCHARTTYPES],
								trChartDataExt ce[NCHARTTYPES], bool do_refresh)
{
	int i;

	for(i = 0; i < NCHARTTYPES; i++)
	{
		chart[i] = ci[i];
		chartdata[i] = cd[i];
		chartdataext[i] = ce[i];
	}

	if(IS_ONE_CHART(mode))
		CalculateAspects(chart1, chart1, FALSE);
	else
		CalculateAspects(chart1, chart2, TRUE);

	if(mode & ASPCONFIG_CHART)
	{
		wxASSERT(IS_ONE_CHART(mode));
		CalculateAspectConfigurations(chart1);
	}
	else
	{
		aspect_conf.Clear();
	}

	infopane->SetChartData(ci, cd, ce);
	infopane->SetRestrictions(restricted, r_aspects);

	if(do_refresh)
		Refresh(FALSE);
}

void CChartCanvas::SetInfoPane(CChartInfo *chartinfopane)
{
	infopane = chartinfopane;
}

void CChartCanvas::SetRestrictions(bool restr[NCHARTTYPES][NTOTOBJECTS+12], bool r_asp[NASPECTS],
								   double asp_orbs[NASPECTS], bool do_refresh)
{
	int i, j;

	for(i = 0; i < NCHARTTYPES; i++)
	{
		for(j = 0; j < (NTOTOBJECTS+12); j++)
		{
			restricted[i][j] = restr[i][j];
		}
	}

	for(i = 0; i < NASPECTS; i++)
	{
		aspect_orb[i] = asp_orbs[i];
		r_aspects[i] = r_asp[i];
	}
	
	if(IS_ONE_CHART(mode))
		CalculateAspects(chart1, chart1, FALSE);
	else
		CalculateAspects(chart1, chart2, TRUE);

	if(mode & ASPCONFIG_CHART)
	{
		wxASSERT(IS_ONE_CHART(mode));
		CalculateAspectConfigurations(mode);
	}
	else
	{
		aspect_conf.Clear();
	}

	infopane->SetRestrictions(restricted, r_aspects, do_refresh);

	if(do_refresh)
		Refresh(FALSE);
}

void CChartCanvas::OnMouseMove(wxMouseEvent &event)
{
    wxClientDC dc(this);
    PrepareDC(dc);
    owner->PrepareDC(dc);

	static bool already_printed[N_ASPCONFS];
	wxCoord center_x, center_y, xx, yy, ww, hh;
	int x1, x2, y1, y2;
	wxString str, longstr, tempstr, tempstr2, hname, str1, str2;
	double r, th, signdeg, dist, linelen;
	trSignPos signpos, signpos2;
	unsigned int k, l;
	int old_mouse_over_obj, old_mouse_over_chart;
	int i, j, n, curobj;
	int n1 = chart1, n2 = chart2;

    wxPoint pos = event.GetPosition();
    wxCoord x = dc.DeviceToLogicalX( pos.x );
    wxCoord y = dc.DeviceToLogicalY( pos.y );

	old_mouse_over_obj   = mouse_over_obj;
	old_mouse_over_chart = mouse_over_chart;

	mouse_over_chart = n1;
	mouse_over_obj   = OVER_NOTHING;
	n_aspects_over   = -1;

	center_x = chartloc[n1].x + chartsize[n1].x/2;
	center_y = chartloc[n1].y + chartsize[n1].y/2;
	r = sqrt((double)(SQ(x - center_x) + SQ(y - center_y)));

	if(mode & ASPCONFIG_CHART)
	{
		for(i = 0; i < N_ASPCONFS; i++)
			already_printed[i] = false;
	}
	
	if(r > r_outer_sign[n1]) // Outside of chart
	{
		str = "";
		if((mouse_over_mode != INFO_MODE_SUMMARY) || (mouse_over_obj != -1))
		{
			mouse_over_mode = INFO_MODE_SUMMARY;
			mouse_over_obj  = -1;
			Refresh(FALSE);
		}
	}
	else if(r > r_inner_sign[n1]) // Signs
	{
			th = atan(((double)y - center_y)/((double)x - center_x));;
			if((x < center_x) && (y < center_y)) // upper-left quadrant
				th = 2.0*N_PI - th;
			else if((x < center_x) && (y > center_y)) // lower-left quadrant
				th = -th;
			else // right half
				th = N_PI - th;
			th *= 180.0/N_PI;

			n = -1;
			for(i = 1; i < 12; i++)
			{
				if(sign[n1][i].deg <= sign[n1][i+1].deg)
				{
					if((th >= sign[n1][i].deg) && (th < sign[n1][i+1].deg))
						n = i;
				}
				else
				{
					if((th >= sign[n1][i].deg) || (th < sign[n1][i+1].deg))
						n = i;
				}
			}
			if(n == -1)
				n = 12;

			str = trGetName(ARIES + sign[n1][n].num - 1);

			if((mouse_over_mode != INFO_MODE_SIGN) ||
				(mouse_over_obj != ARIES + sign[n1][n].num - 1))
			{
				mouse_over_mode = INFO_MODE_SIGN;
				mouse_over_obj  = ARIES + sign[n1][n].num - 1;
				Refresh(FALSE);
			}
	}
	else if(r > r_outer_aspect[n1]) // Planets et al.
	{
		curobj = -1;
		str = "";
		n = -1;
		for(i = 0; i < (NTOTOBJECTS + 12); i++)
		{
			if(restricted[n1][i]) continue;
			xx = obj[n1][i].x;
			yy = obj[n1][i].y;
			ww = obj[n1][i].w;
			hh = obj[n1][i].h;
			if((x >= xx) && (x <= (xx + ww)) && (y >= yy) && (y <= (yy + hh)))
			{
				str = trGetName(obj[n1][i].n);
				mouse_over_mode = INFO_MODE_PLANET;
				mouse_over_obj = obj[n1][i].n;
				n = i;
				break;
			}
		}

		if((IS_TWO_CHARTS(mode)) && (n < 0))
		{
			for(i = 0; i < (NTOTOBJECTS + 12); i++)
			{
				if(restricted[n2][i]) continue;
				xx = obj[n2][i].x;
				yy = obj[n2][i].y;
				ww = obj[n2][i].w;
				hh = obj[n2][i].h;
				if((x >= xx) && (x <= (xx + ww)) && (y >= yy) && (y <= (yy + hh)))
				{
					str = trGetName(obj[n2][i].n);
					mouse_over_chart = n2;
					mouse_over_mode = INFO_MODE_PLANET;
					mouse_over_obj = obj[n2][i].n;
					n = i;
					break;
				}
			}
		}

		if((n >= 0) && (mouse_over_chart == n1))
		{
			signdeg = obj[n1][n].deg + chart[n1].house[1].cusp;
			if(signdeg > 360.0) signdeg -= 360.0;
			split_longitute(signdeg, &signpos, TRUE);
			tempstr.Printf(": %d %s %d", signpos.degrees,
				trGetName(signpos.sign), signpos.minutes);
			
			str += tempstr;

			if((mouse_over_obj != old_mouse_over_obj) ||
				(mouse_over_chart != old_mouse_over_chart))
			{
				Refresh(FALSE);
			}
		}
		else if((n >= 0) && (mouse_over_chart == n2))
		{
			signdeg = obj[n2][n].deg + chart[n1].house[1].cusp;
			if(signdeg > 360.0) signdeg -= 360.0;
			split_longitute(signdeg, &signpos, TRUE);
			tempstr.Printf(": %d %s %d", signpos.degrees,
				trGetName(signpos.sign), signpos.minutes);
			
			str += tempstr;

			if((mouse_over_obj != old_mouse_over_obj) ||
				(mouse_over_chart != old_mouse_over_chart))
			{
				Refresh(FALSE);
			}
		}
		else
		{
			if((mouse_over_obj != old_mouse_over_obj) ||
				(mouse_over_chart != old_mouse_over_chart) ||
				(mouse_over_mode != INFO_MODE_SUMMARY))
			{
				Refresh(FALSE);
			}

			mouse_over_mode = INFO_MODE_SUMMARY;
			mouse_over_obj  = -1;
		}

	}
	else // In the inner circle with the aspects
	{
		str = "";
		longstr = "";
		for(i = 0, n = 0; i < n_aspect_cache; i++)
		{
			aspect_over[i] = false;
		}
		for(i = 0, n = 0; i < n_aspect_cache; i++)
		{
			x1 = aspect_cache[i].x1;
			x2 = aspect_cache[i].x2;
			y1 = aspect_cache[i].y1;
			y2 = aspect_cache[i].y2;
			linelen = sqrt((double)(SQ(x2 - x1) + SQ(y2 - y1)));
			if(linelen > 0.5) // to avoid singularities
				dist = fabs((double)((x2 - x1)*(y1 - y) - (x1 - x)*(y2 - y1)))/linelen;
			else // in the limit that x2 -> x1 and y2 -> y1, estimate the distance as just dx+dy
				dist = sqrt(SQ(y1 - y) + SQ(x1 - x));
			if((dist <= 6) &&
			   (!(mode & ASPCONFIG_CHART) || (aspect_cache[i].aspconf.GetCount() > 0)))
			{
				if(!(mode & ASPCONFIG_CHART))
				{
					// In normal mode, show the aspects the mouse is over
					str1 = trGetAbbrevName(aspect_cache[i].n1);
					str2 = trGetAbbrevName(aspect_cache[i].n2);
					if((mode & TRANSITS_CHART) ||
						(IS_ONE_CHART(mode) && (aspect_cache[i].n1 > aspect_cache[i].n2)))
					{
						tempstr.Printf("%s %s %s | ", trGetAbbrevName(aspect_cache[i].n2),
							trGetAbbrevName(aspect_cache[i].n_aspect),
							trGetAbbrevName(aspect_cache[i].n1));
						tempstr2.Printf("%s %s %s | ", trGetName(aspect_cache[i].n2),
							trGetName(aspect_cache[i].n_aspect),
							trGetName(aspect_cache[i].n1));
					}
					else
					{
						tempstr.Printf("%s %s %s | ", trGetAbbrevName(aspect_cache[i].n1),
							trGetAbbrevName(aspect_cache[i].n_aspect),
							trGetAbbrevName(aspect_cache[i].n2));
						tempstr2.Printf("%s %s %s | ", trGetName(aspect_cache[i].n1),
							trGetName(aspect_cache[i].n_aspect),
							trGetName(aspect_cache[i].n2));
					}
					str += tempstr;
					longstr += tempstr2;
					aspect_over[i] = true;
					n++;
				}
				else // Else if we are in Aspect Configuration mode
				{
					// In Aspect Configuration mode, show both the aspects the
					// mouse is over, and all other aspects in the same
					// configurations
					for(k = 0; k < aspect_cache[i].aspconf.GetCount(); k++)
					{
						for(j = 0; j < n_aspect_cache; j++)
						{
							for(l = 0; l < aspect_cache[j].aspconf.GetCount(); l++)
							{
								if(aspect_cache[i].aspconf[k] == aspect_cache[j].aspconf[l])
								{
									if(!aspect_over[j])
									{
										aspect_over[j] = true;
										n++;
									}
								}
							}
						}
						
						if(!already_printed[aspect_cache[i].aspconf[k]])
						{
							already_printed[aspect_cache[i].aspconf[k]] = true;
							tempstr.Printf("%s | ", trGetAbbrevName(
								aspect_conf[aspect_cache[i].aspconf[k]].kind));
							tempstr2.Printf("%s | ", trGetName(
								aspect_conf[aspect_cache[i].aspconf[k]].kind));
							str += tempstr;
							longstr += tempstr2;
						}
					}
				} // End if in Aspect Configuration mode
			} // End if (dist <= 6)

		}

		if(str.Right(3) == " | ")
			str = str.Mid(0, str.length() - 3);

		if(longstr.Right(3) == " | ")
			longstr = longstr.Mid(0, longstr.length() - 3);

		longstr.Replace(_("Conjunction"), _("Conjunct"));
		longstr.Replace(_("Opposition"), _("Opposing"));

		if(longstr.length() <= 127)
			str = longstr;

		if((n > 0) && (str.length() > 127))
			str.Printf(_("%d aspects near mouse"), n);

		if(n > 0)
		{
			mouse_over_chart = n1;
			mouse_over_mode  = INFO_MODE_ASPECTS;
			mouse_over_obj   = OVER_ASPECTS;
			n_aspects_over   = n;

			Refresh(FALSE);
		}

		if(n <= 0) // If not over any aspects, then select the house instead
		{
			th = atan(((double)y - center_y)/((double)x - center_x));;
			if((x < center_x) && (y < center_y)) // upper-left quadrant
				th = 2.0*N_PI - th;
			else if((x < center_x) && (y > center_y)) // lower-left quadrant
				th = -th;
			else // right half
				th = N_PI - th;
			th *= 180.0/N_PI;
			n = -1;
			for(i = 1; i < 12; i++)
				if((th >= house[n1][i].deg) && (th < house[n1][i+1].deg))
					n = i;
			if(n == -1)
				n = 12;
			switch(n)
			{
			case 1:  hname = _("First");    break;
			case 2:  hname = _("Second");   break;
			case 3:  hname = _("Third");    break;
			case 4:  hname = _("Fourth");   break;
			case 5:  hname = _("Fifth");    break;
			case 6:  hname = _("Sixth");    break;
			case 7:  hname = _("Seventh");  break;
			case 8:  hname = _("Eigth");    break;
			case 9:  hname = _("Ninth");    break;
			case 10: hname = _("Tenth");    break;
			case 11: hname = _("Eleventh"); break;
			case 12: hname = _("Twelfth");  break;
			default: hname = _("Invalid");
			}
			split_longitute(chart[n1].house[n].cusp, &signpos, TRUE);
			split_longitute(chart[n1].house[((n) % 12) + 1].cusp, &signpos2, TRUE);
			str.Printf(_("%s House: %d %s %d - %d %s %d"), hname.c_str(),
				signpos.degrees, trGetName(signpos.sign),
				signpos.minutes,
				signpos2.degrees, trGetName(signpos2.sign),
				signpos2.minutes);
			if((mouse_over_mode != INFO_MODE_HOUSE) ||
				(mouse_over_obj != HOUSE1 + n - 1))
			{
				mouse_over_mode = INFO_MODE_HOUSE;
				mouse_over_obj  = HOUSE1 + n - 1;
				Refresh(FALSE);
			}
		}
	}

    owner->SetStatusText( str );
}

void CChartCanvas::OnMouseDown(wxMouseEvent &event)
{
    wxClientDC dc(this);
    PrepareDC(dc);
    owner->PrepareDC(dc);

	if((infopane->CompareMode(mouse_over_mode, mouse_over_obj, mouse_over_chart)) ||
		(mouse_over_mode == INFO_MODE_ASPECTS))
	{
		if(mouse_over_mode == INFO_MODE_ASPECTS)
		{
			infopane->SetAspectOver(n_aspects_over, aspect_over);
			infopane->SetAspectsMode(0);
		}
		infopane->SetMode(mouse_over_mode, mouse_over_obj, mouse_over_chart);
		infopane->Refresh(FALSE);
	}
}

void CChartCanvas::OnMouseLeaveWindow(wxMouseEvent &event)
{
	mouse_over_mode  = INFO_MODE_SUMMARY;
	mouse_over_obj   = -1;
	mouse_over_chart = chart1;
	owner->SetStatusText("");
}





///////////////////////////////// PAINTING ROUTINES
#define DEGTORAD(a) ((N_PI/180.0)*(a))
#define dlaaMIN_DIST 10

void CChartCanvas::DrawLineAtAngle(wxDC &dc, int x1, int y1, double r1, double r2, 
								   double angle_degrees1, double angle_degrees2, bool is_arrow)
{
	int x2, y2, x3, y3; // The line will go from (x2,y2) to (x3,y3)
	int x4, y4, x5, y5; // The endpoints of the arrow lines (that start from (x3,y3))
	int sgnx1, sgny1, sgnx2, sgny2, quad, dx, dy; // Signs & Quadrant
	double theta1, theta2; // The angular points on the circle of the line endpoints
	double alpha1, alpha2; // Angles used in drawing the arrows
	double theta; // The angle between the x-axis and the slope of the line
	double phi  = DEGTORAD(15.0); // The "spread" of the lines of the arrow
	double alen = 0.085*(r1 + r2)/2.0; // The length of the lines of the arrow
	theta1 = DEGTORAD(180.0 - angle_degrees1);
	theta2 = DEGTORAD(180.0 - angle_degrees2);
	x2 = x1 + iround(r1*cos(theta1));
	y2 = y1 + iround(r1*sin(theta1));
	x3 = x1 + iround(r2*cos(theta2));
	y3 = y1 + iround(r2*sin(theta2));
	/********* TEMP ********/
	/*dc.SetPen(wxPen(wxColor(128,128,128),dc.GetPen().GetWidth(),dc.GetPen().GetStyle()));*/
	/***********************/
	dc.DrawLine(x2, y2, x3, y3);

	dx = abs(x3 - x2);  dy = abs(y3 - y2);
	if(is_arrow && (dx > dlaaMIN_DIST) && (dy > dlaaMIN_DIST) &&
			((SQ(dx) + SQ(dy)) > SQ(alen*2.0)) /*DEBUG*/ && FALSE)
	{
		if(dx != 0) theta = atan(((double)dy)/((double)dx));
		else theta = N_PI/4.0;
		quad = sgnx1 = sgnx2 = sgny1 = sgny2 = 0;
		if     ((x2 >= x3) && (y2 >= y3) && (dx >= dy)) quad = 1;
		else if((x2 >= x3) && (y2 >= y3) && (dx <= dy)) quad = 2;
		else if((x2 <= x3) && (y2 >= y3) && (dx <= dy)) quad = 3;
		else if((x2 <= x3) && (y2 >= y3) && (dx >= dy)) quad = 4;
		else if((x2 <= x3) && (y2 <= y3) && (dx >= dy)) quad = 5;
		else if((x2 <= x3) && (y2 <= y3) && (dx <= dy)) quad = 6;
		else if((x2 >= x3) && (y2 <= y3) && (dx <= dy)) quad = 7;
		else if((x2 >= x3) && (y2 <= y3) && (dx >= dy)) quad = 8;
		wxASSERT(quad != 0);

		

		alpha1 = N_PI/4.0 + theta1/2.0 + theta2/2.0 + phi;
		alpha2 = N_PI/4.0 + theta1/2.0 + theta2/2.0 - phi;
		/*if(alpha1 > N_PI) alpha1 -= N_PI;
		if(alpha1 <    0) alpha1 += N_PI;
		if(alpha2 > N_PI) alpha2 -= N_PI;
		if(alpha2 <    0) alpha2 += N_PI;*/

		if((quad == 1) || (quad == 5))
		{
			alpha1 = theta + phi;
			alpha2 = theta - phi;
			sgnx1 = sgnx2 = sgny1 = sgny2 = (quad == 1) ? 1 : -1;
		}
		else if((quad == 2) || (quad == 6))
		{
			alpha1 = theta + phi;
			alpha2 = theta - phi;
			sgnx1 = sgnx2 = sgny1 = sgny2 = (quad == 2) ? 1 : -1;
		}

		x4 = x3 + iround(sgnx1*alen*acos(alpha1));
		y4 = y3 + iround(sgny1*alen*asin(alpha1));
		x5 = x3 + iround(sgnx2*alen*acos(alpha2));
		y5 = y3 + iround(sgny2*alen*asin(alpha2));
		/********* TEMP ********/
		dc.SetPen(wxPen(wxColor(0,255,255),
			dc.GetPen().GetWidth(),dc.GetPen().GetStyle()));
		if(quad == 6)
				dc.SetPen(wxPen(wxColor(255,128,0),
					dc.GetPen().GetWidth(),dc.GetPen().GetStyle()));
		/*else if(quad == 2) dc.SetPen(*wxCYAN_PEN);
		else if(quad == 3) dc.SetPen(*wxRED_PEN);
		else if(quad == 4) dc.SetPen(*wxGREEN_PEN);
		else dc.SetPen(*wxBLACK_DASHED_PEN);*/
		/***********************/
		if(quad == 1 || quad == 5)
		{
			dc.DrawLine(x3, y3, x4, y4);
			dc.DrawLine(x3, y3, x5, y5);
		}
	}
}

void CChartCanvas::DrawGlyphAtAngle(wxDC &dc, int n, int scale, int x1, int y1,
									double r, double angle_degrees)
{
	DrawGlyph(n, &dc,
		x1 + iround(r*cos(DEGTORAD(180.0 - angle_degrees))),
		y1 + iround(r*sin(DEGTORAD(180.0 - angle_degrees))),
		scale);
}

void CChartCanvas::DrawTextAtAngle(wxDC &dc, int x1, int y1, double r,
								   wxString &text, double angle_degrees)
{
	dc.DrawText(text,
		x1 + iround(r*cos(DEGTORAD(180.0 - angle_degrees))),
		y1 + iround(r*sin(DEGTORAD(180.0 - angle_degrees))));
}

void CChartCanvas::store_loc_radial_deg(int nchart, int nobj, int x1, int y1,
										double r, double angle_degrees,
										int w, int h)
{
	obj[nchart][nobj].x = x1 +
		iround(r*cos(DEGTORAD(180.0 - angle_degrees))) - w/2;
	obj[nchart][nobj].y = y1 +
		iround(r*sin(DEGTORAD(180.0 - angle_degrees))) - h/2;
	obj[nchart][nobj].w = w;
	obj[nchart][nobj].h = h;
}

/*void CChartCanvas::store_aspect(int n_aspect,
								int nchart1, int nchart2, int n1, int n2, int x1, int y1,
								double r1, double r2, double theta1, double theta2, double orb)
{
	aspect_cache[n_aspect_cache].n_aspect = n_aspect;
	aspect_cache[n_aspect_cache].nchart1 = nchart1;
	aspect_cache[n_aspect_cache].nchart2 = nchart2;
	if((n2 >= n1) || (nchart1 != nchart2) || (n_aspect == DISPOSITOR_ASPECT))
	{
		aspect_cache[n_aspect_cache].n1 = n1;
		aspect_cache[n_aspect_cache].n2 = n2;
		aspect_cache[n_aspect_cache].x1 = x1 + r1*cos(DEGTORAD(180.0 - theta1));
		aspect_cache[n_aspect_cache].y1 = y1 + r1*sin(DEGTORAD(180.0 - theta1));
		aspect_cache[n_aspect_cache].x2 = x1 + r2*cos(DEGTORAD(180.0 - theta2));
		aspect_cache[n_aspect_cache].y2 = y1 + r2*sin(DEGTORAD(180.0 - theta2));
	}
	else
	{
		aspect_cache[n_aspect_cache].n1 = n2;
		aspect_cache[n_aspect_cache].n2 = n1;
		aspect_cache[n_aspect_cache].x1 = x1 + r2*cos(DEGTORAD(180.0 - theta2));
		aspect_cache[n_aspect_cache].y1 = y1 + r2*sin(DEGTORAD(180.0 - theta2));
		aspect_cache[n_aspect_cache].x2 = x1 + r1*cos(DEGTORAD(180.0 - theta1));
		aspect_cache[n_aspect_cache].y2 = y1 + r1*sin(DEGTORAD(180.0 - theta1));
	}
	aspect_cache[n_aspect_cache].orb = orb;
	n_aspect_cache++;
}*/

void CChartCanvas::store_aspect(int n_aspect, int nchart1, int nchart2, int obj1, int obj2,
								double theta1, double theta2, double orb, int replacing)
{
	if(replacing < 0) replacing = n_aspect_cache++;
	aspect_cache[replacing].n_aspect = n_aspect;
	aspect_cache[replacing].nchart1 = nchart1;
	aspect_cache[replacing].nchart2 = nchart2;
	aspect_cache[replacing].n1  = obj1;
	aspect_cache[replacing].n2  = obj2;
	aspect_cache[replacing].th1 = theta1;
	aspect_cache[replacing].th2 = theta2;
	aspect_cache[replacing].orb = orb;
}

void CChartCanvas::fixate_aspect(int n, int x1, int y1, double r1, double r2)
{
	// "Fixating" an aspect in the cache means locating where it is
	// to be physically drawn on the screen
	double theta1 = aspect_cache[n].th1;
	double theta2 = aspect_cache[n].th2;

	if((aspect_cache[n].n2 >= aspect_cache[n].n1) ||
	   (aspect_cache[n].nchart1 != aspect_cache[n].nchart2) ||
	   (aspect_cache[n].n_aspect == DISPOSITOR_ASPECT))
	{
		aspect_cache[n].x1 = x1 + iround(r1*cos(DEGTORAD(180.0 - theta1)));
		aspect_cache[n].y1 = y1 + iround(r1*sin(DEGTORAD(180.0 - theta1)));
		aspect_cache[n].x2 = x1 + iround(r2*cos(DEGTORAD(180.0 - theta2)));
		aspect_cache[n].y2 = y1 + iround(r2*sin(DEGTORAD(180.0 - theta2)));
	}
	else
	{
		aspect_cache[n].x1 = x1 + iround(r2*cos(DEGTORAD(180.0 - theta2)));
		aspect_cache[n].y1 = y1 + iround(r2*sin(DEGTORAD(180.0 - theta2)));
		aspect_cache[n].x2 = x1 + iround(r1*cos(DEGTORAD(180.0 - theta1)));
		aspect_cache[n].y2 = y1 + iround(r1*sin(DEGTORAD(180.0 - theta1)));
	}
}

void CChartCanvas::OnResize(wxSizeEvent &event)
{
}

void CChartCanvas::OnPaint(wxPaintEvent &event)
{
	wxCoord w, h, ww, hh, left, top;
	wxPaintDC dc(this);
	PrepareDC(dc);

	GetSize(&w, &h);

	ww = hh = (w > h) ? h : w;
	left  = (w > h) ? ((w - h)/2) : 0;
	top   = (w > h) ? 0 : ((h - w)/2);

	wxBitmap bitbuf(w, h);
	wxMemoryDC memdc;
	memdc.SelectObject(bitbuf);
	memdc.SetBackground(wxBrush(*wxBLACK, wxSOLID));
	memdc.Clear();

	DrawChart(memdc, left, top, ww, hh, chart1, chart2, mode);

	dc.Blit(0, 0, w, h, &memdc, 0, 0);
}

void CChartCanvas::spread_out(int n, double orb)
{
	// This function does its best to make sure that no planets are touching each
	// other on the chart; i.e.: that they are at least orb away from each other
	int ord[NTOTOBJECTS+12+1];
	int i, j, k, tmp, nobjs;
	double deg;
	bool squished_left, squished_right;

	// First sort the objects by degree
	// Use a simple O(N^2) insertion sort. as there aren't that many objects
	nobjs = 0;
	for(i = 0; i < (NTOTOBJECTS + 12); i++)
		if(!restricted[n][i])
			ord[nobjs++] = i;

	for(i = 0; i < nobjs; i++)
	{
		deg = obj[n][ord[i]].deg;
		for(j = i - 1; j >= 0; j--)
		{
			if(obj[n][ord[j]].deg > obj[n][ord[j+1]].deg)
			{
				tmp = ord[j+1];
				ord[j+1] = ord[j];
				ord[j] = tmp;
			}
		}
	}

	ord[nobjs++] = ord[0] + 360;

	for(k = 0; k < (nobjs - 1); k++)
	{
		for(i = 1; i < (nobjs - 1); i++)
		{
			squished_left =  fabs(obj[n][ord[i]].adj_deg - obj[n][ord[i-1]].adj_deg) < orb;
			squished_right = fabs(obj[n][ord[i+1]].adj_deg - obj[n][ord[i]].adj_deg) < orb;
			if(squished_left && squished_right)
			{
				obj[n][ord[i]].adj_deg =
					fmod((obj[n][ord[i-1]].adj_deg + obj[n][ord[i+1]].adj_deg)/2.0 + 360.0, 360.0);
			}
			else if(squished_left)
			{
				obj[n][ord[i]].adj_deg = fmod(obj[n][ord[i-1]].adj_deg + orb + 360.0, 360.0);
			}
			else if(squished_right)
			{
				obj[n][ord[i]].adj_deg = fmod(obj[n][ord[i+1]].adj_deg - orb + 360.0, 360.0);
			}
		}
	}
}

void CChartCanvas::get_chart_dimensions(int w, int h, bool is_duel,
										wxCoord *d, wxCoord *d2,
										int *glyph_scale, int *glyph_size,
										int *glyph_r, int *inner_circle_r)
{
	*d  = (w > h) ? h : w;
	*d  = (int)(0.98*(*d) + 0.5);
	*d2 = (int)(0.78*(*d) + 0.5);
	if((*d - *d2) < 25) *d2 = *d - 25;
	*glyph_scale = ((*d - *d2) < 80) ? 1 : 2;
	*glyph_size = ((*glyph_scale % 2) == 0) ?
		(LARGE_GLYPH_W*(*glyph_scale)/2) : (SMALL_GLYPH_W*(*glyph_scale));
	*glyph_r = ((*d2)/2 >= (8*(*glyph_scale))) ? ((*d2)/2 - (8*(*glyph_scale))) : 0;
	if(!is_duel)
		*inner_circle_r = (*d2)/2 -   14*(*glyph_scale) - int((*d2)/25.0) - 3;
	else
		*inner_circle_r = (*d2)/2 - 2*14*(*glyph_scale) - int((*d2)/25.0) - 3;
}

void CChartCanvas::DrawBareChart(wxDC &dc, int x1, int y1, int w, int h, int n, bool is_duel)
{
	wxCoord d, d2;
	double deg, nextdeg;
	int glyph_scale, glyph_size, glyph_r;
	int inner_circle_r;
	int i, j, x0, y0;
	bool is_sel;

	get_chart_dimensions(w, h, is_duel, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);

	x0 = x1 + w/2;
	y0 = y1 + h/2;

	// Draw the "skelaton" of the chart, with just the circles and spokes
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.SetPen(*wxWHITE_PEN);
	dc.DrawCircle(x0, y0, d/2);
	dc.DrawCircle(x0, y0, d2/2);
	for(i = 1; i <= 12; i++)
	{
		// Draw the houses
		if((i % 3) != 1)
			dc.SetPen(wxPen(wxColour(128,128,128), 1, wxDOT));
		else
			dc.SetPen(wxPen(wxColour(128,128,128), 1,
				(glyph_scale > 1) ? wxLONG_DASH : wxSOLID));
		deg = chart[n].house[i].cusp - chart[n].house[1].cusp;
		if(deg < 0.0) deg += 360.0;
		for(j = 0; j < 8; j++)
			if(fabs(deg + 45.0*j) < 0.1)
				deg = 45*j;
		house[n][i].deg = deg;
		house[n][i].num = i;
		if((i % 3) != 1)
			DrawLineAtAngle(dc, x0, y0, 0, inner_circle_r, deg, deg);
		else
			DrawLineAtAngle(dc, x0, y0, 0, d2/2, deg, deg);

		dc.SetPen(wxPen(wxColour(64,64,64), 1, wxSOLID));
		dc.DrawCircle(x0, y0, inner_circle_r);
		
		nextdeg = chart[n].house[i+1].cusp - chart[n].house[1].cusp;
		if(nextdeg < deg) nextdeg += 360.0;
		is_sel = (mouse_over_mode == INFO_MODE_HOUSE) &&
			(mouse_over_obj == (FIRST_HOUSE + i - 1));
		dc.SetPen(wxPen(
			(!is_sel) ? wxColour(128,128,128) : *wxWHITE,
			1, wxSOLID));
		if(d2 > 200)
		{
			if(i < 10)
			{
				DrawGlyphAtAngle(dc, TURTLE_DRAW_ASCII+TURTLE_NUMBER+i,
					1, x0, y0, 35, fmod((deg+nextdeg)/2.0, 360.0));
			}
			else
			{
				DrawGlyphAtAngle(dc, TURTLE_DRAW_ASCII+TURTLE_NUMBER+(i/10),
					1, x0 - 3, y0, 35, fmod((deg+nextdeg)/2.0, 360.0));
				DrawGlyphAtAngle(dc, TURTLE_DRAW_ASCII+TURTLE_NUMBER+(i%10),
					1, x0 + 3, y0, 35, fmod((deg+nextdeg)/2.0, 360.0));
			}
		}


		// Draw the signs
		dc.SetPen(wxPen(*wxWHITE, 1, wxSOLID));
		deg = 30*(i - 1) - chart[n].house[1].cusp;
		if(deg < 0.0) deg += 360.0;
		sign[n][i].deg = deg;
		sign[n][i].num = i;
		DrawLineAtAngle(dc, x0, y0, d2/2, d/2, deg, deg);

		dc.SetPen(wxPen(DefaultObjectColour(trGetElement(ARIES+i-1), 255),
			 (mouse_over_obj == (ARIES+i-1)) ? 2 : 1, wxSOLID));

		DrawGlyphAtAngle(dc, ARIES+i-1, glyph_scale, x0, y0, (d + d2)/4, deg+15.0);
	}
}

void CChartCanvas::DrawPlanets(wxDC &dc, int x1, int y1, int w, int h,
							   int n, double asc, bool is_duel, bool is_second,
							   bool by_sign, bool by_house)
{
	wxCoord d, d2;
	double deg;
	bool is_sel;
	int glyph_scale, glyph_size, glyph_r;
	int inner_circle_r, line_r;
	int i, x0, y0, penthick, ndis;

	get_chart_dimensions(w, h, is_duel, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);

	if(is_duel && !is_second)
	{
		// Draw the planets of the second chart of the synasty in the interior
		glyph_r -= (glyph_r - inner_circle_r)/2;
	}

	line_r = glyph_r - glyph_size/2 - 3;

	x0 = x1 + w/2;
	y0 = y1 + h/2;

	for(i = 0; i < NTOTOBJECTS; i++)
	{
		// First, prepare by calculating their positions
		if(restricted[n][i])
			continue;
		deg = fmod(chart[n].planet[i].longitude - asc + 360.0, 360.0);
		obj[n][i].deg = deg;
		obj[n][i].adj_deg = deg;
		obj[n][i].n = FIRST_OBJECT + i;
	}
	for(i = 0; i < 12; i++)
	{
		if(restricted[n][NTOTOBJECTS+i])
			continue;
		deg = fmod(chart[n].house[i+1].cusp - asc + 360.0, 360.0);
		obj[n][NTOTOBJECTS+i].deg = deg;
		obj[n][NTOTOBJECTS+i].adj_deg = deg;
		obj[n][NTOTOBJECTS+i].n = FIRST_HOUSE + i;
	}
	spread_out(n, MINSPREAD);

	for(i = 0; i < (NTOTOBJECTS + 12); i++)
	{
		if(restricted[n][i])
			continue;

		if(i < NTOTOBJECTS)
			is_sel = (mouse_over_obj == (FIRST_OBJECT+i)) && (mouse_over_chart == n);
		else
			is_sel = (mouse_over_mode == INFO_MODE_PLANET) &&
				(mouse_over_obj == (FIRST_HOUSE+i-NTOTOBJECTS)) && (mouse_over_chart == n);

		penthick = is_sel ? 2 : 1;

		ndis = -1;
		if(by_sign)
		{
			if(i < NTOTOBJECTS)
				ndis = RulerOf(chart[n].planet[i].sgn.sign) - FIRST_PLANET;
			else
				ndis = RulerOf(chart[n].house[i-NTOTOBJECTS+1].sgn.sign) - FIRST_PLANET;
		}
		else if(by_house)
		{
			if(i < NTOTOBJECTS)
				ndis = RulerOf(chart[n].planet[i].in_house - 1 + ARIES) - FIRST_PLANET;
			else
				ndis = RulerOf((i - NTOTOBJECTS) + ARIES) - FIRST_PLANET;
		}
		if(ndis == i)
			penthick *= 2;

		dc.SetPen(wxPen(DefaultObjectColour(trGetElement(obj[n][i].n), 255),
			 penthick, wxSOLID));
		
		DrawGlyphAtAngle(dc, obj[n][i].n, glyph_scale, x0, y0,
			             glyph_r, obj[n][i].adj_deg);

		if(is_duel && is_second)
			dc.SetPen(wxPen(DefaultObjectColour(trGetElement(obj[n][i].n), 128),
				is_sel ? 2 : 1, wxSOLID));

		DrawLineAtAngle(dc, x0, y0, line_r,
						inner_circle_r + 3, obj[n][i].adj_deg, obj[n][i].deg);
		store_loc_radial_deg(n, i, x0, y0, glyph_r, obj[n][i].adj_deg, glyph_size, glyph_size);

	}
}

void CChartCanvas::CalculateAspects(int n1, int n2, bool is_duel)
{
	double asc, actual_orb, deg1, deg2;
	bool b1, b2, is_connected;
	int maxj, nobj1, nobj2, replacing;
	int i, j, k, l, n_a;

	asc = chart[n1].house[1].cusp;
	n_aspect_cache = 0;
	for(i = 0; i < (NTOTOBJECTS + 12); i++)
	{
		if(restricted[n1][i]) continue;
		// the second term of the next loop condition is to prevent house-house
		// aspects, like the silly Asc/Des opposition in every chart
		if(n1 == n2)
			maxj = (i < NTOTOBJECTS) ? i : NTOTOBJECTS;
		else
			maxj = NTOTOBJECTS + 12;
		for(j = 0; j < maxj; j++)
		{
			if(restricted[n2][j]) continue;
			for(k = 0; k < NASPECTS; k++)
			{
				if(r_aspects[k]) continue;

				if(i < NTOTOBJECTS)
					deg1 = fmod(chart[n1].planet[i].longitude - asc + 360.0, 360.0);
				else
					deg1 = fmod(chart[n1].house[i-NTOTOBJECTS+1].cusp - asc + 360.0, 360.0);
				if(j < NTOTOBJECTS)
					deg2 = fmod(chart[n2].planet[j].longitude - asc + 360.0, 360.0);
				else
					deg2 = fmod(chart[n2].house[j-NTOTOBJECTS+1].cusp - asc + 360.0, 360.0);
 				n_a = k + FIRST_ASPECT;
				if((j == (MARS - FIRST_PLANET)) &&(i == (JUPITER - FIRST_PLANET)))
					n_a = k + FIRST_ASPECT - 0;
				b1 = fabs(fabs(deg1 - deg2) - AspectDef(n_a)) <
						aspect_orb[n_a - FIRST_ASPECT];
				b2 = fabs((360.0 - fabs(deg1 - deg2)) - AspectDef(n_a)) <
						aspect_orb[n_a - FIRST_ASPECT];
				if(b1)
					actual_orb = fabs(fabs(deg1 - deg2) - AspectDef(n_a));
				else if(b2)
					actual_orb = fabs((360.0 - fabs(deg1 - deg2)) - AspectDef(n_a));
				else
					actual_orb = 180.0; // to avoid compiler warnings
				if(b1 || b2)
				{
					nobj1 = (i < NTOTOBJECTS) ? (FIRST_OBJECT+i) : (FIRST_HOUSE+i-NTOTOBJECTS);
					nobj2 = (j < NTOTOBJECTS) ? (FIRST_OBJECT+j) : (FIRST_HOUSE+j-NTOTOBJECTS);
					is_connected = FALSE;
					replacing = -1;
					for(l = 0; (l < n_aspect_cache) && !is_connected; l++)
					{
						is_connected = is_connected ||
							((aspect_cache[l].n1 == nobj1) && (aspect_cache[l].n2 == nobj2) &&
							 (aspect_cache[l].nchart1 == n1) && (aspect_cache[l].nchart2 == n2));
						// Aspects with lower orbs take priority
						if(is_connected && (actual_orb < aspect_cache[l].orb))
							replacing = l;
					}
					if((!is_connected) || (replacing >= 0))
						store_aspect(n_a, n1, n2, nobj1, nobj2, deg1, deg2,
						             actual_orb, replacing);
				}
			}
		}
	}

	if( infopane != NULL ) infopane->SetAspects(n_aspect_cache, aspect_cache);
}

void CChartCanvas::CalculateAspectConfigurations(int n)
{
	// Watch the parade of the variety of integer variables ;)
	wxArrayInt asps, aspectsleft[N_ASPCONFS];
	const int *a[N_ASPCONFS];
	size_t max_pl, max_depth;
	bool outofrunning[N_ASPCONFS];
	int i, j;

	aspect_conf.Clear();
	for(i = 0; i < n_aspect_cache; i++)
	{
		aspect_cache[i].aspconf.Clear();
		aspect_cache[i].aspconf.Shrink();
	}

	// max_pl is the maximum number of planets in a configuration,
	// and max_depth is the maximum number of aspects needed to be
	// checked, which also controls the recursion depth
	max_pl = max_depth = 0;
	for(i = 0; i < N_ASPCONFS; i++)
	{
		a[i] = GetAspConfigurationDef(i + FIRST_ASPCONF);
		wxASSERT(a[i] != NULL);
		if(((unsigned int)a[i][0]) > max_pl) max_pl = (unsigned int)a[i][0];
		if(((unsigned int)a[i][1]) > max_depth) max_depth = (unsigned int)a[i][1];
	}

	// Initialize aspectsleft with *all* of the aspects for each
	// aspect configuration, and make sure none are out of the running
	// in the beginning
	for(i = 0; i < N_ASPCONFS; i++)
	{
		outofrunning[i] = false;
		for(j = 0; j < a[i][1]; j++)
			aspectsleft[i].Add(a[i][j + 2]);
	}

	asps.Clear();
	asps.Add(0);
	aspect_conf_find(asps, a, max_pl, max_depth, aspectsleft, outofrunning);
}

// Recursive function called by CalculateAspectConfigurations
void CChartCanvas::aspect_conf_find(wxArrayInt &asps, const int *aspconfdef[],
									unsigned int max_pl, unsigned int max_depth,
									wxArrayInt aspectsleft[], bool outofrunning[])
{
	// Do a breadth-first recursive search for aspect configurations.
	// Base Case: No 1 aspect can ever make an aspect configuration
	// Recursion: See if adding the nth aspect will make a valid
	//            aspect configuration.
	// Prune the recursion tree by seeing if, given the current aspects,
	// no aspect configuration is possible anymore.  Also, stop at the
	// maximum recursion depth, and if too many planets are involved.
	//
	// All the pruning makes this much more efficient than nested for()
	// loops on the planets, even though it's more complicated to
	// implement.  Because of its complexity, I've tried to comment it
	// more thoroughly than the rest of my code. --mikesag, 7/7/2004

	

	// Make the maximum number of variables static so they won't have
	// to be instantiated every time this function is recursively called
	wxArrayInt pl;
	static aspect_configuration *new_aspconf;
	static unsigned int j, k;
	static int q, val, loc;
	static bool allcantbe;
	bool cantbe[N_ASPCONFS]; // Is a copy of outofrunning()
	bool made_cantbe_here[N_ASPCONFS], dont_recurse;
	unsigned int n_asps, n_pl;
	int  removed[N_ASPCONFS];
	int  r, n1, n2;
#ifdef __WXDEBUG__
	static int n_recs = 0; // For benchmarking/profiling purposes
	n_recs++;
#endif

	// Determine the number of planets involved so far
	pl.Clear();
	n_asps = asps.GetCount();
	for(k = 0; k < n_asps; k++)
	{
		n1 = aspect_cache[asps[k]].n1;
		n2 = aspect_cache[asps[k]].n2;
		if(pl.Index(n1) == wxNOT_FOUND) pl.Add(n1);
		if(pl.Index(n2) == wxNOT_FOUND) pl.Add(n2);
	}
	n_pl = pl.GetCount();

	// If we've gone over the maximum number of planets in all
	// aspect configurations, stop the recursion.
	// Value of pruning the tree like this with the standard
	// 12-planet configuration, with 6 maximum planets, on
	// three different dates/times/places:
	//       523,127 vs.  9,085 calls to this function
	//     1,042,379 vs.  7,175 calls to this function
	//    15,505,589 vs. 81,618 calls to this function
	// This should show the value of pruning the search
	// tree whenever humanly (or computerly) possible
	if(n_pl > max_pl)
		return;

	// r is the "recursion" variable; we store the last element of the
	// array here, pushing an incremented value onto the asps array
	// when recursing, and restoring it upon return
	r = asps[asps.GetCount() - 1];
	
	// The first recursive step goes "sideways" so to speak; it adds
	// no more length to the aspect array we're using in the search,
	// but instead increments the last element of the array, keeping
	// the number of elements the same.

	// Only recurse if we haven't gone over the total number of
	// aspects in the chart.
	if((r + 1) < n_aspect_cache)
	{
		// First, replace the last value of the array with an
		// incremented value and recurse
		asps.RemoveAt(asps.GetCount() - 1);
		asps.Add(r + 1);
		aspect_conf_find(asps, aspconfdef, max_pl, max_depth,
			aspectsleft, outofrunning);
		asps.RemoveAt(asps.GetCount() - 1);
		asps.Add(r);
	}

	// Now, remove this aspect from all configurations that have it.
	// (We use an array of arrays to keep track of which aspect
	// configurations are still possible, and which aspects still need
	// to occur before we have a complete aspect configuration.)
	// If one doesn't have it, it's out of the running.  If, after removal,
	// one has no aspects left in it, then it's an aspect configuration!
	// When we either have a complete aspect configuration, or we
	// encounter an aspect that's not in a particular aspect configuration,
	// take it out of the running of possible aspect configurations for
	// this particular set of aspects.  (Backtracking might bring it back
	// though, so we have to keep track of what we've done to the
	// aspect configuration candidate array-of-arrays at each step.)
	// This allows a spectacular performance increase when we stop the
	// recursion if we find that no aspect configurations are possible
	// after taking the latest aspect into account, as we'll see later.
	allcantbe = true;
	val = aspect_cache[asps.Last()].n_aspect;
	for(q = 0; q < N_ASPCONFS; q++)
	{
		removed[q] = -1;
		made_cantbe_here[q] = false;
		cantbe[q] = outofrunning[q];
		if(cantbe[q])
		{
			continue;
		}

		if((loc = aspectsleft[q].Index(val)) == wxNOT_FOUND)
		{
			made_cantbe_here[q] = true;
			cantbe[q] = true;
			continue;
		}

		removed[q] = val;
		aspectsleft[q].RemoveAt(loc);

		if(aspectsleft[q].GetCount() == 0)
		{
			made_cantbe_here[q] = true;
			cantbe[q] = true;
			
			if(((int)n_pl) == aspconfdef[q][0])
			{
				// We've found an aspect-configuration!
				new_aspconf = new aspect_configuration;
				new_aspconf->kind = FIRST_ASPCONF + q;
				for(j = 0; j < n_pl; j++)
				{
					new_aspconf->planet.Add(pl[j]);
				}
				for(j = 0; j < n_asps; j++)
				{
					new_aspconf->aspect.Add(asps[j]);
					aspect_cache[asps[j]].aspconf.Add(aspect_conf.GetCount());
				}
				aspect_conf.Add(new_aspconf);
			}

			continue;
		}

		allcantbe = false;
	}
	
	// One more pruning of the tree: If none of the aspect configurations are
	// possible anymore (according to cantbe[]), then stop the recursion;
	// there's no point.  With the same test (with the origional number of
	// calls to this function first, followed by the number of calls after
	// the pruning of too many planets, followed by the combination of that
	// and the pruning when no more configurations are possible given the
	// aspects so far):
	//       523,127 vs.  9,085 vs.  1,630 calls to this function
	//     1,042,379 vs.  7,175 vs.  2,319 calls to this function
	//    15,505,589 vs. 81,618 vs. 10,125 calls to this function
	dont_recurse = false;
	if(allcantbe)
		dont_recurse = true;

	// Only recurse if we haven't gone over maximum number of aspects
	// This time, we are going deeper into the recursion; adding one
	// more aspect onto the list.
	if(((r + 1) < n_aspect_cache) && !dont_recurse)
	{
		// If we haven't reached the maximum recursion depth yet,
		// recurse one level deeper by adding the incremented value
		// onto the end of the array
		if(asps.GetCount() < max_depth)
		{
			asps.Add(r + 1);
			aspect_conf_find(asps, aspconfdef, max_pl, max_depth,
				aspectsleft, cantbe);
			asps.RemoveAt(asps.GetCount() - 1);
		}
	}

	// Restore what has been changed for the calling instance of
	// the function
	for(q = 0; q < N_ASPCONFS; q++)
	{
		// Note that order doesn't matter for aspectsleft[]
		if(made_cantbe_here[q])
		{
			made_cantbe_here[q] = FALSE;
			cantbe[q] = FALSE;
		}
		if(removed[q] >= 0) aspectsleft[q].Add(removed[q]);
	}

// To show a lot of debugging information, leave the next line
// uncommented; otherwise, comment it out
#define __VERBOSE
#ifdef __WXDEBUG__
	if((asps.GetCount() == 1) && (asps[0] == 0))
	{
#ifdef __VERBOSE
		wxString s, s2;
		s.Printf("Number of times recursed: %d", n_recs);

		s += "\n\nPlanets in Aspect Configurations:\n";
		for(q = 0; q < (int)aspect_conf.GetCount(); q++)
		{
			s2.Printf("%d: ", q);
			s += s2;
			for(r = 0; r < (int)aspect_conf[q].planet.GetCount(); r++)
			{
				s2.Printf("%s ", trGetAbbrevName(aspect_conf[q].planet[r]));
				s += s2;
			}
			s2.Printf(" (%s)\n", trGetName(aspect_conf[q].kind));
			s += s2;
		}

		s += "\n\nAspects in Aspect Configurations:\n";
		for(q = 0; q < (int)aspect_conf.GetCount(); q++)
		{
			s2.Printf("%d: ", q);
			s += s2;
			for(r = 0; r < (int)aspect_conf[q].aspect.GetCount(); r++)
			{
				s2.Printf("%d ", aspect_conf[q].aspect[r]);
				s += s2;
			}
			s2.Printf(" (%s)\n", trGetName(aspect_conf[q].kind));
			s += s2;
		}

		s += "\n\nAspect Configuration Links in Aspect Cache:\n";
		for(q = 0; q < n_aspect_cache; q++)
		{
			s2.Printf("%d (%s-%s): ", q, trGetAbbrevName(aspect_cache[q].n1),
				trGetAbbrevName(aspect_cache[q].n2));
			s += s2;
			for(r = 0; r < (int)aspect_cache[q].aspconf.GetCount(); r++)
			{
				s2.Printf("%d ", aspect_cache[q].aspconf[r]);
				s += s2;
			}
			s += "\n";
		}
		wxMessageBox(s, "Aspect Configuration Debug");
#endif
		n_recs = 0;
	}
#endif
}

void CChartCanvas::DrawAspects(wxDC &dc, int x1, int y1, int w, int h,
							   int n1, int n2, bool is_duel, bool is_aspconfig)
{
	wxCoord d, d2;
	double asp_stren, intens, deg1, deg2;
	int glyph_scale, glyph_size, glyph_r;
	int penstyle, penthick;
	int inner_circle_r;
	int i, n_a, x0, y0;

	get_chart_dimensions(w, h, is_duel, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);

	x0 = x1 + w/2;
	y0 = y1 + h/2;

	dc.SetPen(wxPen(*wxCYAN, 1, wxSOLID));
	for(i = 0; i < n_aspect_cache; i++)
	{
		if(is_aspconfig && !(aspect_cache[i].aspconf.GetCount() > 0))
			continue;
		n_a = aspect_cache[i].n_aspect;
		deg1 = aspect_cache[i].th1;
		deg2 = aspect_cache[i].th2;
		penstyle = wxSOLID;
		penthick = (n_a == CONJUNCTION) ? 2 : 1;
		if(mouse_over_mode == INFO_MODE_ASPECTS)
		{
			if(aspect_over[i])
				penthick *= 2;
		}
		asp_stren = (aspect_orb[n_a - FIRST_ASPECT] - aspect_cache[i].orb)/
			aspect_orb[n_a - FIRST_ASPECT];
		if(!(mode & ASPCONFIG_CHART))
			intens = (asp_stren > 0) ?
				((int)floor(256*asp_stren + (0.5/256.0))) : 0;
		else
			intens = (asp_stren > 0) ?
				((int)floor(128*asp_stren + (0.5/128.0)) + 128) : 0;
		dc.SetPen(wxPen(DefaultAspectColour(
			n_a, (unsigned char)iround(intens)), penthick, penstyle));
		DrawLineAtAngle(dc, x0, y0, inner_circle_r, inner_circle_r, deg1, deg2);
		fixate_aspect(i, x0, y0, inner_circle_r, inner_circle_r);
	}
}

void CChartCanvas::DrawDispositors(wxDC &dc, int x1, int y1, int w, int h, int n, bool by_sign)
{
	wxColour col1, col2, col3;
	wxCoord d, d2;
	double deg1, deg2;
	int glyph_scale, glyph_size, glyph_r;
	int inner_circle_r, penthick, ndis;
	int i, x0, y0;

	get_chart_dimensions(w, h, FALSE, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);

	x0 = x1 + w/2;
	y0 = y1 + h/2;

	n_aspect_cache = 0;
	for(i = 0; i < (NTOTOBJECTS + 12); i++)
	{
		if(restricted[n][i]) continue;

		if(by_sign)
		{
			if(i < NTOTOBJECTS)
				ndis = RulerOf(chart[n].planet[i].sgn.sign) - FIRST_PLANET;
			else
				ndis = RulerOf(chart[n].house[i-NTOTOBJECTS+1].sgn.sign) - FIRST_PLANET;
		}
		else
		{
			if(i < NTOTOBJECTS)
				ndis = RulerOf(chart[n].planet[i].in_house - 1 + ARIES) - FIRST_PLANET;
			else
				ndis = RulerOf((i - NTOTOBJECTS) + ARIES) - FIRST_PLANET;
		}

		deg1 = obj[n][i].deg;
		deg2 = obj[n][ndis].deg;		

		penthick = ((mouse_over_mode == INFO_MODE_ASPECTS) &&
			aspect_over[n_aspect_cache]) ? 2 : 1;
		col1 = DefaultObjectColour(trGetElement(i+FIRST_PLANET), 255);
		col2 = DefaultObjectColour(trGetElement(ndis+FIRST_PLANET), 255);
		col3.Set((col1.Red() + col2.Red())/2, (col1.Green() + col2.Green())/2,
			(col1.Blue() + col2.Blue())/2);
		/* dc.SetPen(wxPen(DefaultAspectColour(DISPOSITOR_ASPECT, 255), penthick, wxSOLID)); */
		dc.SetPen(wxPen(col3, penthick, wxSOLID));
		DrawLineAtAngle(dc, x0, y0, inner_circle_r, inner_circle_r, deg1, deg2, TRUE);
		store_aspect(DISPOSITOR_ASPECT, n, n, obj[n][i].n, obj[n][ndis].n, deg1, deg2, 0);
		fixate_aspect(n_aspect_cache, x0, y0, inner_circle_r, inner_circle_r);
	}

	if( infopane != NULL ) infopane->SetAspects(n_aspect_cache, aspect_cache);
}

void CChartCanvas::DrawChart(wxDC &dc, int x1, int y1, int w, int h, int n1,
							 int n2, int chart_kind)
{
	// To the chart drawer, a composite chart is just another kind of natal chart
	// We have to know about it to pass it to the info pane, however.  Ditto with
	// Davison charts.
	if(IS_ONE_CHART(chart_kind))
		DrawNatalChart(dc, x1, y1, w, h, n1, chart_kind);
	// Ditto with synasty and transit charts
	else if(IS_TWO_CHARTS(chart_kind))
		DrawSynastyChart(dc, x1, y1, w, h, n1, n2, chart_kind);
}

void CChartCanvas::DrawNatalChart(wxDC &dc, int x1, int y1, int w, int h,
								  int n, int chart_kind)
{
	wxCoord d, d2;
	int glyph_scale, glyph_size, glyph_r;
	int inner_circle_r;

	get_chart_dimensions(w, h, FALSE, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);
	
	chartloc[n].x     = x1;
	chartloc[n].y     = y1;
	chartsize[n].x    = w;
	chartsize[n].y    = h;
	r_outer_sign[n]   = d/2;
	r_inner_sign[n]   = d2/2;
	r_outer_aspect[n] = inner_circle_r;

	// Draw the chart outline, signs, and houses
	DrawBareChart(dc, x1, y1, w, h, n, FALSE);

	// Draw Planets & Other Objects
	DrawPlanets(dc, x1, y1, w, h, n, chart[n].house[1].cusp, FALSE, FALSE,
		(chart_kind & DISPOSITORS_CHART) != 0, (chart_kind & DISPOSITORS_HOUSE_CHART) != 0);

	// Draw Aspects
	if(chart_kind & ASPECTS_CHART)
		DrawAspects(dc, x1, y1, w, h, n, n, FALSE, FALSE);
	else if(chart_kind & ASPCONFIG_CHART)
		DrawAspects(dc, x1, y1, w, h, n, n, FALSE, TRUE);
	else if(chart_kind & DISPOSITORS_CHART)
		DrawDispositors(dc, x1, y1, w, h, n, TRUE);
	else if(chart_kind & DISPOSITORS_HOUSE_CHART)
		DrawDispositors(dc, x1, y1, w, h, n, FALSE);
	else
		n_aspect_cache = 0;
}

void CChartCanvas::DrawSynastyChart(wxDC &dc, int x1, int y1, int w, int h,
									int n1, int n2, int chart_kind)
{
	wxCoord d, d2;
	int glyph_scale, glyph_size, glyph_r;
	int inner_circle_r;

	get_chart_dimensions(w, h, TRUE, &d, &d2, &glyph_scale, &glyph_size,
	                     &glyph_r, &inner_circle_r);
	
	chartloc[n1].x     = x1;
	chartloc[n1].y     = y1;
	chartsize[n1].x    = w;
	chartsize[n1].y    = h;
	r_outer_sign[n1]   = d/2;
	r_inner_sign[n1]   = d2/2;
	r_outer_aspect[n1] = inner_circle_r;

	// Draw the chart outline, signs, and houses
	DrawBareChart(dc, x1, y1, w, h, n1, TRUE);

	// Draw Planets & Other Objects
	DrawPlanets(dc, x1, y1, w, h, n1, chart[n1].house[1].cusp, TRUE, FALSE, FALSE, FALSE);
	DrawPlanets(dc, x1, y1, w, h, n2, chart[n1].house[1].cusp, TRUE, TRUE, FALSE, FALSE);

	// Draw Aspects
	if(chart_kind & ASPECTS_CHART)
		DrawAspects(dc, x1, y1, w, h, n1, n2, TRUE);
	else
		n_aspect_cache = 0;
}
