#include "wx/wxprec.h"

#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

#include <math.h>
#include <locale.h>

#include "wx/msgdlg.h"
#include "wx/string.h"
#include "wx/tokenzr.h"
#include "wx/dynarray.h"
#include "wx/filename.h"
#include "wx/textfile.h"
#include "wx/ffile.h"

#include "Translation.h"
#include "TransitsFrm.h"
#include "TransitsFiles.h"

#if defined(__WXMSW__) && defined(WIN32)
#include <windows.h>
#endif

#define NATIVE_HEADER_LINE    "Transitarium Data File, Format Version: 1.0"
#define NATIVE_HEADER_STANZA  "Transitarium Data File, Format Version"
#define NATIVE_CURRENT_VERSION 1.0

#define CONFIG_HEADER_LINE    "Transitarium Configuration File, Format Version: 1.0"
#define CONFIG_HEADER_STANZA  "Transitarium Configuration File, Format Version"
#define CONFIG_CURRENT_VERSION 1.0

// ----------------------------------------------------------------------------
// Configuration File Routines
// ----------------------------------------------------------------------------

static bool InterpretRestrictionValue(const wxString &str)
{
	if((str.CmpNoCase("X") == 0) || (str.CmpNoCase(_("No")) == 0) ||
			(str.CmpNoCase(_("Restrict")) == 0) ||
			(str.CmpNoCase(_("Restricted")) == 0) ||
			(str.CmpNoCase(_("Non")) == 0))
		return true;

	return false;
}

static const char *RestrictionToString(bool is_restricted)
{
	return (is_restricted ? "X" : "-");
}

static const char *BoolToString(bool b)
{
	return (b ? "True" : "False");
}

bool TransitsFiles::IsValidConfigFile(const wxString &filename, bool verbose)
{	
	wxFFile raw_f;
	wxTextFile f;
	wxString headerstr, stanza, verstr;
	double ver;
	char header[14];

	if(!raw_f.Open(filename))
	{
		if(verbose)
			wxMessageBox("Error opening file!", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		return false;
	}
	
	if(raw_f.Length() < (strlen(CONFIG_HEADER_STANZA) + 2))
	{
		if(verbose)
			wxMessageBox("Too small of a file", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}

	raw_f.Read((void *)header, 12); // Should be "Transitarium" ;)
	header[12] = '\0';
	if(wxString(CONFIG_HEADER_STANZA).Mid(0, 12).CmpNoCase(header) != 0)
	{
		if(verbose)
			wxMessageBox("Invalid Chart Data File", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}

	raw_f.Close();
	f.Open(filename);
	headerstr = f.GetFirstLine();
	f.Close();

	headerstr = headerstr.BeforeFirst('\n');
	headerstr = headerstr.BeforeFirst('#');
	stanza = headerstr.BeforeFirst(':');
	verstr = headerstr.AfterFirst(':');
	stanza.Trim(TRUE);
	stanza.Trim(FALSE);
	verstr.Trim(TRUE);
	verstr.Trim(FALSE);
	verstr.ToDouble(&ver);
	if((stanza.CmpNoCase(CONFIG_HEADER_STANZA) != 0) || (ver < 0.01))
	{
		if(verbose)
			wxMessageBox(_("Invalid Chart Data File"), wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		return false;
	}
	if(ver > CONFIG_CURRENT_VERSION)
	{
		if(verbose)
		{
			if(wxMessageBox(_("Too new a version of configuration file for this \
version of the program.  Try opening anyway?"),
				wxMessageBoxCaptionStr,
				wxYES_NO | wxCENTER | wxICON_QUESTION) == wxYES)
					return true;
		}

		return false;
	}
	
	return true;
}

void TransitsFiles::GetConfigFile(TransitsFrame *tfr, int nchart,
								  const wxString &filename, wxUint32 flags)
{
	wxString line, var, valstr, s;
	wxTextFile f;
	int i;
	double dbl;
	size_t lc;

	if(!IsValidConfigFile(filename, TRUE))
		return;
	
	f.Open(filename);
	for(lc = 0; lc < f.GetLineCount(); lc++)
	{
		line = f.GetLine(lc);
		
		// All lines are in a "variable: value #comment" format, which
		// makes it easy to parse and hopefully also easy to read.

		// First, remove the comment
		line = line.BeforeFirst('#');

		// Then, split into variable and value sections, and remove
		// the surrounding whitespace from both
		var = line.BeforeFirst(':');
		valstr = line.AfterFirst(':');
		var.Trim(TRUE);
		var.Trim(FALSE);
		valstr.Trim(TRUE);
		valstr.Trim(FALSE);

		// Most likely, a blank line, or a comment line; continue on
		// to the next line in the case there is no ':'
		if(line.Cmp(var) == 0)
			continue;

		// Now interpret each one!
		if((flags & trCONFIG_CALC_SETTINGS) || (flags & trCONFIG_ALL))
		{
			// Various settings, all with their own formats
			if((var.CmpNoCase("House System") == 0) ||
				(var.CmpNoCase("HouseSys") == 0))
			{
				for(i = 0; i < NHSYSTEMS; i++)
				{
					if(valstr.CmpNoCase(trGetName(FIRST_HSYSTEM + i)) == 0)
						tfr->hsys = FIRST_HSYSTEM + i;
				}
			}
			else if((var.CmpNoCase("Moon's Node Type") == 0) ||
				(var.CmpNoCase("North Node Type") == 0))
			{
				if(valstr.Lower().Contains("mean"))
					tfr->nodemode = USE_MEAN;
				else
					tfr->nodemode = USE_TRUE;
				set_node_type(tfr->nodemode);
			}
			else if((var.CmpNoCase("Apogee Type") == 0) ||
				(var.CmpNoCase("Lilith Type") == 0))
			{
				if(valstr.Lower().Contains("true"))
					tfr->apogmode = USE_TRUE;
				else
					tfr->apogmode = USE_MEAN;
				set_apog_type(tfr->apogmode);
			}
		}
		if((flags & trCONFIG_PLANETS_SHOWN) || (flags & trCONFIG_ALL))
		{
			// Example: Mercury: -
			for(i = 0; i < NTOTOBJECTS; i++)
			{
				if(var.CmpNoCase(trGetName(FIRST_OBJECT + i)) == 0)
					tfr->restricted[nchart][i] =
						InterpretRestrictionValue(valstr);
			}
			for(i = 0; i < 12; i++)
			{
				if(var.CmpNoCase(trGetName(FIRST_HOUSE + i)) == 0)
					tfr->restricted[nchart][NTOTOBJECTS+i] =
						InterpretRestrictionValue(valstr);
			}
		}
		if((flags & trCONFIG_ASPECTS_SHOWN) || (flags & trCONFIG_ALL))
		{
			// Example: Sextile: X
			for(i = 0; i < NASPECTS; i++)
			{
				if(var.CmpNoCase(trGetName(FIRST_ASPECT + i)) == 0)
					tfr->r_aspects[i] =
						InterpretRestrictionValue(valstr);
			}
		}
		if((flags & trCONFIG_ASPECT_ORBS) || (flags & trCONFIG_ALL))
		{
			// Example: Sextile Orb: 6.25
			for(i = 0; i < NASPECTS; i++)
			{
				s = trGetName(FIRST_ASPECT + i);
				s += " Orb";
				valstr.ToDouble(&dbl);
				if(var.CmpNoCase(s) == 0)
					tfr->aspect_orb[i] = dbl;
			}
		}
		if((flags & trCONFIG_ANIM_SETTINGS) || (flags & trCONFIG_ALL))
		{
			if(var.CmpNoCase("Separate Chart Speeds") == 0)
			{
				tfr->TimeMotionDlg->sep_timers->SetValue(
					valstr.Lower().Contains("true"));
			}
			else if(var.CmpNoCase("Smooth Movement") == 0)
			{
				tfr->TimeMotionDlg->smooth_movement->SetValue(
					valstr.Lower().Contains("true"));
			}
			else if(var.CmpNoCase("Pause at Each Stop") == 0)
			{
				tfr->TimeMotionDlg->pause_toggle->SetValue(
					valstr.Lower().Contains("true"));
			}
			else if(var.CmpNoCase("Pause Duration") == 0)
			{
				valstr.ToDouble(&dbl);
				if(dbl < 0.000000001) dbl = 2.5;
				if(fabs(dbl - int(floor(dbl + 0.5))) < 0.0001)
					s.Printf("%d", int(floor(dbl + 0.5)));
				else
					s.Printf("%.1f", dbl);
				tfr->TimeMotionDlg->pause_time_text->SetValue(s);
			}
			else if(var.CmpNoCase("Frame Rate") == 0)
			{
				valstr.ToDouble(&dbl);
				if(dbl < 1.0) dbl = 1.0;
				if(dbl > 100.0) dbl = 100.0;
				tfr->TimeMotionDlg->frame_rate_slider->SetValue(
					(int)floor(dbl + 0.5));
			}	
			else if(var.CmpNoCase("Step Method") == 0)
			{
				for(i = 0; i < tfr->TimeMotionDlg->step_method->GetCount(); i++)
				{
					if(valstr.CmpNoCase(
						tfr->TimeMotionDlg->step_method->GetString(i)) == 0)
					{
						tfr->TimeMotionDlg->step_method->SetSelection(i);
					}
				}
			}
		}
	}
	f.Close();
		
	if(tfr->nowtimer->IsRunning())
		tfr->nowtimer->Stop();

	tfr->Recalculate();
	tfr->SetChartsToView(tfr->mode);
	tfr->ChartCanvas->SetRestrictions(tfr->restricted, tfr->r_aspects, tfr->aspect_orb);
	tfr->ChartCanvas->SetMode(tfr->mode | tfr->viewmode,
		tfr->chart0, tfr->chart1, tfr->chart2, FALSE, FALSE);
	tfr->ChartCanvas->SetChartData(tfr->chart, tfr->chartdata,
		tfr->chartdataext, TRUE);
	tfr->TimeMotionDlg->UpdateDialog();
	tfr->TimeMotionDlg->UpdateSpans();

	if(tfr->Animator.IsRunning())
		tfr->Animator.Restart();
}

void TransitsFiles::PutConfigFile(TransitsFrame *tfr, int nchart,
								  const wxString &filename, wxUint32 flags,
								  bool overwrite)
{
	int i, nobj;
	unsigned int maxlen;
	bool is_new_file;
	wxTextFile f, tmpf;
	wxString str, str2;
	wxFileName tmpfilename;
	
	tmpfilename.AssignTempFileName("");

	is_new_file = !wxFileExists(filename);
	if(!is_new_file && !overwrite)
		f.Open(filename);
	
	if(!tmpf.Open(tmpfilename.GetFullPath()))
	{
		wxMessageBox("Error opening file " + filename,
			wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
		return;
	}
	
	// Note that it's the version of the file format, not the program!
	tmpf.AddLine(CONFIG_HEADER_LINE);
	tmpf.AddLine("");
	
	if(is_new_file || overwrite)
	{
		if((flags & trCONFIG_CALC_SETTINGS) || (flags & trCONFIG_ALL))
		{
			str.Printf("%s: %s", "House System", trGetName(tfr->hsys));
			tmpf.AddLine(str);

			str.Printf("%s: %s", "North Node Type",
				(tfr->nodemode == USE_MEAN) ? "Mean" : "True");
			tmpf.AddLine(str);

			str.Printf("%s: %s", "Lilith Type",
				(tfr->apogmode == USE_TRUE) ? "True" : "Mean");
			tmpf.AddLine(str);

			tmpf.AddLine("");
		}
		if((flags & trCONFIG_PLANETS_SHOWN) || (flags & trCONFIG_ALL))
		{
			// Example: Mercury: -
			for(i = 0, maxlen = 0; i < (NTOTOBJECTS + 12); i++)
			{
				nobj = (i < NTOTOBJECTS) ? (FIRST_OBJECT + i) :
					(FIRST_HOUSE + i - NTOTOBJECTS);
				if(strlen(trGetName(nobj)) > maxlen)
					maxlen = strlen(trGetName(nobj));
			}
			for(i = 0; i < (NTOTOBJECTS + 12); i++)
			{
				nobj = (i < NTOTOBJECTS) ? (FIRST_OBJECT + i) :
					(FIRST_HOUSE + i - NTOTOBJECTS);
				str.Printf("%s: %*s", trGetName(nobj),
					maxlen - strlen(trGetName(nobj)) + 1,
					RestrictionToString(tfr->restricted[nchart][i]));
				tmpf.AddLine(str);
			}
			tmpf.AddLine("");
		}
		if((flags & trCONFIG_ASPECTS_SHOWN) || (flags & trCONFIG_ALL))
		{
			// Example: Sextile: X
			for(i = 0, maxlen = 0; i < NASPECTS; i++)
			{
				if(strlen(trGetName(FIRST_ASPECT + i)) > maxlen)
					maxlen = strlen(trGetName(FIRST_ASPECT + i));
			}
			for(i = 0; i < NASPECTS; i++)
			{
				str.Printf("%s: %*s", trGetName(FIRST_ASPECT + i),
					maxlen - strlen(trGetName(FIRST_ASPECT + i)) + 1,
					RestrictionToString(tfr->r_aspects[i]));
				tmpf.AddLine(str);
			}
			tmpf.AddLine("");
		}
		if((flags & trCONFIG_ASPECT_ORBS) || (flags & trCONFIG_ALL))
		{
			// Example: Sextile Orb: 6.25
			for(i = 0, maxlen = 0; i < NASPECTS; i++)
			{
				if(strlen(trGetName(FIRST_ASPECT + i)) > maxlen)
					maxlen = strlen(trGetName(FIRST_ASPECT + i));
			}
			for(i = 0; i < NASPECTS; i++)
			{
				str2.Printf("%s", trGetName(FIRST_ASPECT + i));
				str2 += " Orb";
				str.Printf("%s: %*s%.2lf", str2.c_str(),
					maxlen - strlen(trGetName(FIRST_ASPECT + i)), "",
					tfr->aspect_orb[i]);
				tmpf.AddLine(str);
			}
			tmpf.AddLine("");
		}
		if((flags & trCONFIG_ANIM_SETTINGS) || (flags & trCONFIG_ALL))
		{
			str.Printf("%s: %s", "Separate Chart Speeds",
				BoolToString(tfr->TimeMotionDlg->sep_timers->GetValue()));
			tmpf.AddLine(str);

			str.Printf("%s: %s", "Smooth Movement",
				BoolToString(tfr->TimeMotionDlg->smooth_movement->GetValue()));
			tmpf.AddLine(str);

			str.Printf("%s: %s", "Pause at Each Step",
				BoolToString(tfr->TimeMotionDlg->pause_toggle->GetValue()));
			tmpf.AddLine(str);

			str.Printf("%s: %s", "Pause Duration",
				tfr->TimeMotionDlg->pause_time_text->GetValue().c_str());
			tmpf.AddLine(str);

			str.Printf("%s: %d", "Frame Rate",
				tfr->TimeMotionDlg->frame_rate_slider->GetValue());
			tmpf.AddLine(str);

			str.Printf("%s: %s", "Step Method",
				tfr->TimeMotionDlg->step_method->GetStringSelection().c_str());
			tmpf.AddLine(str);

			tmpf.AddLine("");
		}
	}
	
	if(!is_new_file && !overwrite)
		f.Close();
	tmpf.Write();
	tmpf.Close();

	if(!is_new_file)
	{
		if(!wxRemoveFile(filename))
		{
			wxMessageBox("Error removing existing file " + filename,
				wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
			return;
		}
	}
	
	if( !wxRenameFile(tmpfilename.GetFullPath(), filename) )
	{
		wxMessageBox("Error moving " + tmpfilename.GetFullPath() +
			" to " + filename,
			wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
		return;
	}
}

// ----------------------------------------------------------------------------
// Native Chart Data File Routines
// ----------------------------------------------------------------------------

bool TransitsFiles::IsValidChartDataFile(const wxString &filename, bool verbose)
{	
	wxFFile raw_f;
	wxTextFile f;
	wxString headerstr, stanza, verstr;
	double ver;
	char header[14];

	if(!raw_f.Open(filename))
	{
		if(verbose)
			wxMessageBox("Error opening file!", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		return false;
	}
	
	if(raw_f.Length() < (strlen(NATIVE_HEADER_LINE) + 2))
	{
		if(verbose)
			wxMessageBox("Too small of a file", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}

	raw_f.Read((void *)header, 12); // Should be "Transitarium" ;)
	header[12] = '\0';
	if(wxString(NATIVE_HEADER_STANZA).Mid(0, 12).CmpNoCase(header) != 0)
	{
		if(verbose)
			wxMessageBox("Invalid Chart Data File", wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}

	raw_f.Close();
	f.Open(filename);
	headerstr = f.GetFirstLine();
	f.Close();

	headerstr = headerstr.BeforeFirst('\n');
	headerstr = headerstr.BeforeFirst('#');
	stanza = headerstr.BeforeFirst(':');
	verstr = headerstr.AfterFirst(':');
	stanza.Trim(TRUE);
	stanza.Trim(FALSE);
	verstr.Trim(TRUE);
	verstr.Trim(FALSE);
	verstr.ToDouble(&ver);
	if((stanza.CmpNoCase(NATIVE_HEADER_STANZA) != 0) || (ver < 0.01))
	{
		if(verbose)
			wxMessageBox(_("Invalid Chart Data File"), wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		return false;
	}
	if(ver > NATIVE_CURRENT_VERSION)
	{
		if(verbose)
		{
			if(wxMessageBox(_("Too new a version of chart data file for this \
version of the program.  Try opening anyway?"),
				wxMessageBoxCaptionStr,
				wxYES_NO | wxCENTER | wxICON_QUESTION) == wxYES)
					return true;
		}
		return false;
	}
	
	return true;
}

void TransitsFiles::GetChartDataFile(TransitsFrame *tfr, int nchart,
									 const wxString &filename)
{
	wxString line, var, valstr, valstr2, nowtm, nowdt;
	bool is_date_now, are_both_now, is_dst_now;
	wxDateTime wxdt;
	wxTextFile f;
	size_t i;
	long int n;

	wxdt = wxDateTime::Now();
	nowtm = wxdt.FormatTime();
	nowdt = wxdt.FormatDate();
	are_both_now = false;
	is_date_now = false;
	is_dst_now = wxdt.IsDST() ? true : false;

	// If the user tries to open an Astrolog file as a native file,
	// silently import it instead of giving an error.
	// Better to guess the user's mind than be a nitpicking nanny
	// in this case.  (Best not to try *too* hard though, for that
	// would frustrate more expert users too much.)
	if(IsValidAstrologFile(filename, FALSE))
	{
		GetAstrologFile(tfr, nchart, filename);
		return;
	}

	if(!IsValidChartDataFile(filename, TRUE))
		return;
	
	f.Open(filename);
	for(i = 0; i < f.GetLineCount(); i++)
	{
		line = f.GetLine(i);
		
		// All lines are in a "variable: value #comment" format, which
		// makes it easy to parse and hopefully also easy to read.

		// First, remove the comment
		line = line.BeforeFirst('#');

		// Then, split into variable and value sections, and remove
		// the surrounding whitespace from both
		var = line.BeforeFirst(':');
		valstr = line.AfterFirst(':');
		var.Trim(TRUE);
		var.Trim(FALSE);
		valstr.Trim(TRUE);
		valstr.Trim(FALSE);

		// Most likely, a blank line, or a comment line; continue on
		// to the next line in the case there is no ':'
		if(line.Cmp(var) == 0)
			continue;

		// Now interpret each one!
		if(var.CmpNoCase("Name") == 0)
		{
			tfr->chartdataext[nchart].namestr = valstr;
		}
		else if(var.CmpNoCase("Date") == 0)
		{
			if(valstr.CmpNoCase("Now") == 0)
			{
				is_date_now = true;
				StringToTrDate(nowdt, &(tfr->chartdata[nchart].date));
				tfr->chartdata[nchart].is_progressed =
					(nchart == CHART_DEFAULT) ? TRUE : FALSE;
				tfr->chartdataext[nchart].set_to_now = true;
			}
			else
			{
				is_date_now = false;
				StringToTrDate(valstr, &(tfr->chartdata[nchart].date));
				tfr->chartdata[nchart].is_progressed = FALSE;
				tfr->chartdataext[nchart].set_to_now = false;
			}
		}
		else if(var.CmpNoCase("Time") == 0)
		{
			if(!is_date_now)
			{
				if(valstr.CmpNoCase("Now") == 0)
					StringToHMS(nowtm, &(tfr->chartdata[nchart].time));
				else
					StringToHMS(valstr, &(tfr->chartdata[nchart].time));
			}
			else // both data and time are set to Now
			{
				are_both_now = true;
			}

		}
		else if(var.CmpNoCase("Timezone") == 0)
		{
			// Break it up into the time zone itself, and then
			// the possible DST marker
			valstr2 = valstr.AfterFirst(' ');
			valstr = valstr.BeforeFirst(' ');
			valstr.Trim(TRUE);
			valstr2.Trim(FALSE);
			valstr2.MakeUpper();

			valstr.ToLong(&n);
			tfr->chartdata[nchart].timezone = (int)n;

			if(!is_date_now)
				tfr->chartdata[nchart].is_dst = valstr2.Contains("DST");
			else
				tfr->chartdata[nchart].is_dst = is_dst_now;
		}
		else if(var.CmpNoCase("City") == 0)
		{
			tfr->chartdataext[nchart].citystr = valstr;
		}
		else if(var.CmpNoCase("Coordinates") == 0)
		{
			// Break it up into longitude and latitude
			valstr2 = valstr.AfterFirst(' ');
			valstr = valstr.BeforeFirst(' ');
			valstr.Trim(TRUE);
			valstr2.Trim(FALSE);
			valstr2.MakeUpper();

			// Assume longitude then latitude; if error, try the reverse
			if(!StringToDMS(valstr, &(tfr->chartdata[nchart].longitude), "E", "W", TRUE))
			{
				StringToDMS(valstr2, &(tfr->chartdata[nchart].latitude), "N", "S", FALSE);
			}
			else
			{
				StringToDMS(valstr, &(tfr->chartdata[nchart].latitude), "N", "S", FALSE);
				StringToDMS(valstr2, &(tfr->chartdata[nchart].longitude), "E", "W", TRUE);
			}
		}
	}
	f.Close();
	
	if(are_both_now)
		tfr->SetToNow(nchart, false, false);

	tfr->chartdataext[nchart].set_to_now = are_both_now;
	tfr->is_now = are_both_now;
	if(nchart == CHART_DEFAULT)
		tfr->chartdata[nchart].is_progressed = are_both_now;
	
	if(tfr->nowtimer->IsRunning())
		tfr->nowtimer->Stop();

	tfr->Recalculate();
	tfr->SetChartsToView(tfr->mode);
	tfr->ChartCanvas->SetMode(tfr->mode | tfr->viewmode,
		tfr->chart0, tfr->chart1, tfr->chart2, FALSE, FALSE);
	tfr->ChartCanvas->SetChartData(tfr->chart, tfr->chartdata,
		tfr->chartdataext, TRUE);
	if(tfr->Animator.IsRunning())
		tfr->Animator.Restart();

}

void TransitsFiles::PutChartDataFile(TransitsFrame *tfr, int nchart,
									 const wxString &filename)
{
	bool is_now;
	wxTextFile f;
	wxString str, str2, str3;

	if(wxFileExists(filename))
	{
		if(!wxRemoveFile(filename))
		{
			wxMessageBox("Error removing existing file " + filename,
				wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
			return;
		}
	}
	
	if(!f.Create(filename))
	{
		wxMessageBox("Error creating file " + filename,
			wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
		return;
	}

	// Determine whether we should put in the date and time of the
	// chart as entered, or Now.  In general, use Now if and only if:
	//   (a) we're writing the default chart and the is_progressed
	//       flag is set, or
	//   (b) we're writing a chart other than the default chart, and
	//       the chart has been set to "Now" and never changed
	is_now = ((nchart == CHART_DEFAULT) && tfr->chartdata[nchart].is_progressed)
		|| ((nchart != CHART_DEFAULT) && (tfr->chartdataext[nchart].set_to_now));
	
	// Note that it's the version of the file format, not the program!
	f.AddLine(NATIVE_HEADER_LINE);
	f.AddLine("");
	
	str.Printf("Name: %s", tfr->chartdataext[nchart].namestr.c_str());
	f.AddLine(str);

	// Although we use the locale to display times and dates, store
	// the times and dates in a (hopefully) non-controversial
	// locale-independent format for compatability purposes.
	// This means month names in English only!  (Unless the program
	// itself has been translated.)
	str.Printf("Date: %d MONTH %d", tfr->chartdata[nchart].date.day,
		tfr->chartdata[nchart].date.year);
	str.Replace("MONTH",
		wxString(trGetName(JANUARY+tfr->chartdata[nchart].date.month-1)));
	if(!is_now)
		f.AddLine(str);
	else
		f.AddLine("Date: Now");

	// That said, use 12 or 24-hour as locale-appropriate; neither
	// is particularly hard to parse (or read) by anybody
	str = "Time: " + HMStoString(&(tfr->chartdata[nchart].time));
	if(!is_now)
		f.AddLine(str);
	else
		f.AddLine("Time: Now");
	
	// Now, for the time zone (and DST information)
	str.Printf("Timezone: %d%s # %s",
		tfr->chartdata[nchart].timezone,
		(tfr->chartdata[nchart].is_dst && !is_now) ? " (DST)" : "",
		"Positive values are Eastern hemisphere; negative, Western");
	f.AddLine(str);

	str.Printf("City: %s", tfr->chartdataext[nchart].citystr.c_str());
	f.AddLine(str);

	str2 = DMStoString(&(tfr->chartdata[nchart].longitude), "E", "W");
	str3 = DMStoString(&(tfr->chartdata[nchart].latitude),  "N", "S");
	str.Printf("Coordinates: %s %s", str2.c_str(), str3.c_str());
	f.AddLine(str);
	
	f.Write();
	f.Close();
}



// ----------------------------------------------------------------------------
// Astrolog File Routines
// ----------------------------------------------------------------------------

bool TransitsFiles::IsValidAstrologFile(const wxString &filename, bool verbose)
{
	wxFFile raw_f;
	char header[6];
	header[5] = '\0';

	if(!raw_f.Open(filename))
	{
		if(verbose)
			wxMessageBox(_("Error opening file!"), wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		return false;
	}
	
	if(raw_f.Length() < 10)		
	{
		if(verbose)
			wxMessageBox(_("Too small of a file"), wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}
	
	raw_f.Read((void *)header, 5);
	if(strcmp(header, "@0102") != 0)
	{
		if(verbose)
			wxMessageBox(_("Invalid Astrolog File"), wxMessageBoxCaptionStr,
				wxOK | wxCENTER | wxICON_ERROR);
		raw_f.Close();
		return false;
	}
	
	raw_f.Close();

	return true;
}

void TransitsFiles::GetAstrologFile(TransitsFrame *tfr, int nchart,
									const wxString &filename)
{
	wxString line, tok;
	wxStringTokenizer toks, tokstime;
	wxTextFile f;
	long int n;
	size_t i;

	if(!IsValidAstrologFile(filename, TRUE))
		return;
	
	f.Open(filename);
	//for(line = f.GetFirstLine(); !f.Eof(); line = f.GetNextLine())
	for(i = 0; i < f.GetLineCount(); i++)
	{
		line = f.GetLine(i);
		if(line.substr(0,3) == "/qb") // chart data line
		{
			toks.SetString(line, " ");
			
			// throw out the "/qb" up front
			tok = toks.GetNextToken();
			
			// next up: the date in month, day, year format
			tok  = toks.GetNextToken();
			tok += " " + toks.GetNextToken();
			tok += " " + toks.GetNextToken();
			StringToTrDate(tok, &(tfr->chartdata[nchart].date));
			
			// now the time
			tok = toks.GetNextToken();
			StringToHMS(tok, &(tfr->chartdata[nchart].time));
			
			// daylight savings time?
			tok = toks.GetNextToken();
			tfr->chartdata[nchart].is_dst = (tok == "DT");
			
			// time zone
			tok = toks.GetNextToken();
			tokstime.SetString(tok, ":");
			tok = tokstime.GetNextToken();
			tok.ToLong(&n);
			tfr->chartdata[nchart].timezone = (int)(-n);
			
			// longitude and latitude
			tok = toks.GetNextToken();
			StringToDMS(tok, &(tfr->chartdata[nchart].longitude), "E", "W", TRUE);
			
			tok = toks.GetNextToken();
			StringToDMS(tok, &(tfr->chartdata[nchart].latitude), "N", "S", FALSE);

			// progressed flag
			tfr->chartdata[nchart].is_progressed = FALSE;
			tfr->chartdataext[nchart].set_to_now = false;
		}
		else if(line.substr(0,3) == "/zi") // name and location line
		{
			toks.SetString(line, "\"");
			
			// throw out the "/zi" up front
			tok = toks.GetNextToken();

			// get first the name, then the city
			tok = toks.GetNextToken();
			tfr->chartdataext[nchart].namestr = tok;
			tok = toks.GetNextToken();
			tok = toks.GetNextToken();
			tfr->chartdataext[nchart].citystr = tok;
		}
	}
	f.Close();
	
	tfr->is_now = false;
	tfr->chartdataext[nchart].set_to_now = false;
	
	if(tfr->nowtimer->IsRunning())
		tfr->nowtimer->Stop();

	tfr->Recalculate();
	tfr->SetChartsToView(tfr->mode);
	tfr->ChartCanvas->SetMode(tfr->mode | tfr->viewmode,
		tfr->chart0, tfr->chart1, tfr->chart2, FALSE, FALSE);
	tfr->ChartCanvas->SetChartData(tfr->chart, tfr->chartdata,
		tfr->chartdataext, TRUE);

	if(tfr->Animator.IsRunning())
		tfr->Animator.Restart();
}

void TransitsFiles::PutAstrologFile(TransitsFrame *tfr, int nchart,
									const wxString &filename)
{
	wxTextFile f;
	wxString str, str2;

	if(wxFileExists(filename))
	{
		if(!wxRemoveFile(filename))
		{
			wxMessageBox("Error removing existing file " + filename,
				wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
			return;
		}
	}
	
	if(!f.Create(filename))
	{
		wxMessageBox("Error creating file " + filename,
			wxMessageBoxCaptionStr, wxOK | wxICON_ERROR, tfr);
		return;
	}
	
	f.AddLine("@0102  ; Astrolog chart info.");
	str.Printf("/qb MONTH %d %d %2d:%02d:%02d ?T ",
		tfr->chartdata[nchart].date.day,
		tfr->chartdata[nchart].date.year, // Date (w/o month)
		tfr->chartdata[nchart].time.hours,
		tfr->chartdata[nchart].time.min, // Time
		tfr->chartdata[nchart].time.sec
		);
	str.Replace("MONTH",
		(wxChar *)(wxString(trGetName(JANUARY+tfr->chartdata[nchart].date.month-1)).substr(0,3)).c_str(), false);
	str.Replace("?", tfr->chartdata[nchart].is_dst ? "D" : "S");
	str2.Printf("%+d:00 %3d:%02d:%02dN",
		-(tfr->chartdata[nchart].timezone), // Time Zone
		abs(tfr->chartdata[nchart].longitude.deg),
		abs(tfr->chartdata[nchart].longitude.min), // Longitude
		(int)floor(fabs(tfr->chartdata[nchart].longitude.sec) + 0.5)
		);
	str2.Replace("N",
		is_negative_dms(&(tfr->chartdata[nchart].longitude)) ? "W" : "E" , FALSE);
	str += str2;
	str2.Printf(" %2d:%02d:%02dT",
		abs(tfr->chartdata[nchart].latitude.deg),
		abs(tfr->chartdata[nchart].latitude.min), // Latitude
		(int)floor(fabs(tfr->chartdata[nchart].latitude.sec) + 0.5)
		);
	str2.Replace("T",
		is_negative_dms(&(tfr->chartdata[nchart].latitude))  ? "S" : "N" , FALSE);
	str += str2;
	f.AddLine(str);
	str.Printf("/zi \"%s\" \"%s\"", tfr->chartdataext[nchart].namestr.c_str(),
		tfr->chartdataext[nchart].citystr.c_str());
	f.AddLine(str);
	
	f.Write();
	f.Close();
}

// ----------------------------------------------------------------------------
// utility functions
// ----------------------------------------------------------------------------

static int chartoint(char c)
{
	if(c == '0')
		return 0;
	else if((c >= '1') || (c <= '9'))
		return (c - '1' + 1);

	return 0;
}

wxString DMStoString(const trDMS *dms, const wxString &pos,
					 const wxString &neg)
{
	wxString mystr;
	mystr.Printf("%d:%02d:%02dHEMI",
			abs(dms->deg), abs(dms->min), (int)(fabs(dms->sec) + 0.5)
		);
	mystr.Replace("HEMI", is_negative_dms(dms) ? neg : pos , TRUE);
	return mystr;
}

// Returns non-zero on failure
int StringToDMS(const wxString &str, trDMS *dms,
				const wxString &pos, const wxString &neg, bool max180)
{
	long n;
	int dg, mn, place, almost, succeeded, failed, ispos;
	size_t i;
	wxChar ch, chpos, chneg;
	wxStringTokenizer tokssplit;
	wxString tok;

	// First, try to interpret it as xxx[NSEW]yy
	dg = mn = place = succeeded = failed = almost = ispos = 0;
	chpos = pos.GetChar(0);
	chneg = neg.GetChar(0);
	for(i = 0; (i < str.length()) && !succeeded && !failed; i++)
	{
		ch = str.GetChar(i);
		if((place == 0) && !failed)
		{
			if(isdigit(ch))
				dg = 10*dg + chartoint(ch);
			else if(isalpha(ch))
				place = 1;
			else
				failed = 1;
		}
		if((place == 1) && !failed)
		{
			if((toupper(ch) == chpos) || (tolower(ch) == chpos))
				ispos = 1;
			else if((toupper(ch) == chneg) || (tolower(ch) == chneg))
				ispos = 0;
			else
				failed = 1;
			place = 2;
			continue;
		}
		if((place == 2) && !failed)
		{
			almost = 1;
			if(isdigit(ch))
				mn = 10*mn + chartoint(ch);
			else
				succeeded = 1;
		}
	}
	if(almost) succeeded = 1;
	if(succeeded)
	{
		if(max180)
			dms->deg = dg % 180;
		else
			dms->deg = dg % 90;
		dms->min = mn % 60;
		dms->sec = 0.0;
		if(!ispos)
			negate_dms(dms);
		return 0;
	}

	// If that doesn't work, try the traditional xxx:yy:zz[NSEW] form
	tokssplit.SetString(str, ":");
	tok = tokssplit.GetNextToken();
	if(tok.length() == 0) return 1;
	tok.ToLong(&n);
	dms->deg = n % (max180 ? 180 : 90);
	tok = tokssplit.GetNextToken();
	if(tok.length() == 0) return 1;
	tok.ToLong(&n);
	dms->min = n % 60;
	if(tokssplit.HasMoreTokens())
	{
		tok = tokssplit.GetNextToken();
		tok.ToLong(&n);
		dms->sec = (double)(n % 60);
	}
	else
		dms->sec = 0.0;
	// assume positive unless explicitly negative
	if(tok.Right(neg.length()).CmpNoCase(neg) == 0)
		negate_dms(dms);
	else if(tok.Right(neg.length()).CmpNoCase(pos) != 0)
		return 1;

	return 0;
}

wxString trDateToString(const trDate *trdt)
{
	//TEMP -- TO DELETE (old version)
	//wxString str;
	//str.Printf("MONTH %d %d", trdt->day, trdt->year);
	//str.Replace("MONTH",
	//	wxString(trGetName(JANUARY+(trdt->month)-1)).substr(0,3));
	//return str;

	wxString str;
	char s[256];
	tm tp;

	tp.tm_wday = 0;
	tp.tm_mday = trdt->day;
	tp.tm_mon  = trdt->month - 1;
	tp.tm_year = trdt->year - 1900;

	setlocale(LC_ALL, "");
	strftime(s, 255, "%d %B %Y", &tp);

	str = s;
	if(s[0] == '0')
		str = str.Right(str.length() - 1);
	
	return str;
}

void StringToTrDate(const wxString &str, trDate *trdt)
{
	wxDateTime dati;
	dati.ParseDate(str);
	trdt->day   = dati.GetDay();
	trdt->month = (dati.GetMonth() - wxDateTime::Jan) + 1;
	trdt->year  = dati.GetYear();
}

wxString HMStoString(const trHMS *hms)
{
	wxString str;
	char s[256];
	tm tp;

	tp.tm_hour = hms->hours;
	tp.tm_min  = hms->min;
	tp.tm_sec  = (int)floor(hms->sec);

	setlocale(LC_ALL, "");
	strftime(s, 255, "%X", &tp);

	str = s;
	if((s[0] == '0') &&
		((str.Find('a', 0) >= 0) || (str.Find('A', 0) >= 0) || 
		 (str.Find('p', 0) >= 0) || (str.Find('P', 0) >= 0)))
	{
		str = str.Right(str.length() - 1);
	}
	
	return str;
}

void StringToHMS(const wxString &str, trHMS *hms)
{
	wxDateTime dati;
	dati.ParseTime(str);
	hms->hours = dati.GetHour();
	hms->min   = dati.GetMinute();
	hms->sec   = dati.GetSecond();
}

// Make sure filenames are 8.3 on Win16, and make more "beautiful"
// names on other platforms

const wxString trPrefsDefChart(const wxString &progdir)
{
	return (trPrefsDir(progdir) + wxFileName::GetPathSeparator() +
#if defined(__WXMSW__) && !defined(WIN32)
		"default.trs");
#elif defined(__WXMSW__)
		"Default Chart.trs");
#else // UN*X
		"default_chart.trs");
#endif
}

const wxString trPrefsFile(const wxString &progdir)
{
	return (trPrefsDir(progdir) + wxFileName::GetPathSeparator() +
#if defined(__WXMSW__) && !defined(WIN32)
		"default.trc");
#elif defined(__WXMSW__)
		"Default Configuration.trc");
#else // UN*X
		"default_config.trc");
#endif
}

// This should be correct, unless running MacOS 8 or 9.
// Mac programmers, please help!
const wxString trPrefsDir(const wxString &progdir)
{
#ifdef __WXMSW__
	wxString s;

	switch(GetWindowsVersion())
	{
	case trWIN32S: return progdir;
	case trWIN9X:  return progdir;
	case trWINNT: 
		if(wxGetEnv("AppData", &s))
			return (s + "\\Transitarium");
		else
			return progdir;
	}
	return wxString("C:\\Transitarium");
#else // UN*X-like operating-system
	wxFileName fnm("~/.transitarium");

	fnm.Normalize(wxPATH_NORM_ALL);
	return fnm.GetFullPath();
#endif
}

// The Windows version is needed to determine where to store the
// configuration files.  They will be stored in the program
// folder on Win9x (or, theoretically, in Win3.x), and in
// "%AppData%\Transitarium" on WinNT/2K/XP.
#ifdef __WXMSW__
#ifdef WIN32 // Windows 9x or above
int GetWindowsVersion()
{
	int version = 0;
	OSVERSIONINFO osviVersionInfo;
	
	osviVersionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osviVersionInfo);
	switch(osviVersionInfo.dwPlatformId)
	{
	case VER_PLATFORM_WIN32s:
		version = trWIN32S;
		break;
	case VER_PLATFORM_WIN32_NT:
		version = trWINNT;
		break;
	case VER_PLATFORM_WIN32_WINDOWS:
		version = trWIN9X;
		break;
	default:
		version = trWIN9X;
#ifdef __WXDEBUG__
		wxMessageBox("Error; can't determine the version of Windows used!");
#endif
	}

	return version;
}
#else // MS Windows 3.x; ugh!  (Sorry, no support for MS Windows 2 or below!)
int GetWindowsVersion() { return trWIN32S; }
#endif
#endif
