//Typewriter.cpp
//
//Copyright (C) 2003 mikesag
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "wx/wxprec.h"

#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

//#include "wx/settings.h"
#include "wx/tokenzr.h"

#include <math.h>

#include "AstroConst.h"
#include "AstroCalc.h"

#include "TurtleDraw.h"
#include "Translation.h"

#include "Typewriter.h"

static const int DEFAULT_FAMILY = wxMODERN;
static const int NO_GLYPH = -8756;

// ----------------------------------------------------------------------------
// typewriter drawing implementaton
// ----------------------------------------------------------------------------

CTypewriterDraw::CTypewriterDraw()
{
	x = y = x0 = y0 = 0;
	glsize = 10;
}

void CTypewriterDraw::Reset(int xmargain, int ymargain, int totw, int toth)
{
	x = x0 = xmargain;
	y = y0 = ymargain;
	w = totw;
	h = toth;
}

void CTypewriterDraw::SetMargains(int xmargain, int ymargain)
{
	x = xmargain;
	y = ymargain;
}

void CTypewriterDraw::NewLine(int size)
{
	y += size;
	x = x0;
}

void CTypewriterDraw::Resposition(int newx, int newy)
{
	x = newx;  // warning; this can be left of the margain!
	y = newy;
}

void CTypewriterDraw::TypeText(wxDC &dc, const wxString &s, const wxColour &col,
							   int ptsize, bool bold, int just, int family)
{
	int n_newlines = 0;
	int ww, hh;
	size_t ii;
	wxString str;
	wxFont the_font = wxFont(ptsize, family, wxNORMAL,
	                         bold ? wxBOLD : wxNORMAL, FALSE);

	if(s.IsEmpty())
	{
#ifdef __WXDEBUG__
		wxMessageBox("Warning: Empty string passed to TypeText()", "WARNING!",
			wxICON_WARNING | wxCENTER);
#endif
		return;
	}

	for(ii = s.length() - 1; ii >= 0; ii--)
		if(s[ii] == '\n') n_newlines++;
		else break;

	str = s.substr(0, s.length() - n_newlines);

	dc.SetFont(the_font);
	dc.SetTextForeground(col);
	dc.GetTextExtent(str, &ww, &hh, NULL, NULL, &the_font);
	switch(just)
	{
	case wxRIGHT:
		dc.DrawText(str, (w - x0) - ww, y);
		x = (w - x0) - ww;
		break;
	case wxCENTER:
		dc.DrawText(str, (w - ww)/2, y);
		x = (w - ww)/2;
		break;
	default:
		dc.DrawText(str, x, y);
	}
	if(n_newlines <= 0)
	{
		x += ww;
	}
	else
	{
		x = x0;
		y += n_newlines*hh;
	}
}

void CTypewriterDraw::TypeGlyph(wxDC &dc, int n, const wxColour &col,
								int scale, int thickness, bool centered,
								bool newline_at_end)
{
	const int glscaled = glsize*scale;
#if defined(__WXMSW__)
	const int fudge_x = 0, fudge_y = 3;
#elif defined(__WXGTK__)
	const int fudge_x = 0, fudge_y = ((scale > 1) ? 2 : 0);
#else
	const int fudge_x = 0, fudge_y = 0;
#endif

	if(n != NO_GLYPH)
	{
		dc.SetPen(wxPen(col, thickness, wxSOLID));
		if(!centered)
			DrawGlyph(n, &dc, x + glscaled/2 + fudge_x*scale,
			y + glscaled/2 + fudge_y*scale, scale);
		else
			DrawGlyph(n, &dc, w/2 + fudge_x*scale,
			y + glscaled/2 + fudge_y*scale, scale);
	}

	if(!newline_at_end)
	{
		if(!centered)
			x += glscaled;
		else
			x = (w + glscaled)/2;
	}
	else
	{
		x = x0;
		y += glscaled + fudge_y*scale;
	}
}

int CTypewriterDraw::GetGlyphLineHeight(int scale)
{
#if defined(__WXMSW__)
	const int fudge_y = 3;
#elif defined(__WXGTK__)
	const int fudge_y = ((scale > 1) ? 2 : 0);
#else
	const int fudge_y = 0;
#endif

	return glsize*scale + fudge_y*scale;
}

int CTypewriterDraw::TextExtentX(wxDC &dc, const wxString &s, int ptsize,
								 bool bold, int family)
{
	size_t ii;
	int ww, hh;
	int n_newlines = 0;
	wxString str;
	wxFont the_font = wxFont(ptsize, family, wxNORMAL,
	                         bold ? wxBOLD : wxNORMAL, FALSE);

	for(ii = s.length() - 1; ii >= 0; ii--)
		if(s[ii] == '\n') n_newlines++;
		else break;

	str = s.substr(0, s.length() - n_newlines);

	dc.SetFont(the_font);
	dc.GetTextExtent(str, &ww, &hh, NULL, NULL, &the_font);

	return ww;
}

void CTypewriterDraw::TypeSignPos(wxDC &dc, int n, double longitude, double speed,
								  int glscale, int textscale,
								  int align, bool newline_at_end)
{
	wxColour textcol;
	wxString str;
	trSignPos signpos;

	split_longitute(longitude, &signpos, FALSE);

	str.Printf("/s%d/%2d /s%d//g%d//s%d/ %02d'%02d\"",
		textscale, signpos.degrees, glscale, signpos.sign,
		textscale, signpos.minutes, (int)floor(signpos.seconds + 0.5));
	// Mark retrograde planets
	// (The True Node normally is retrograde, so only mark when it's direct)
	if(speed < 0.0)
		str += "r";
	else if((n == NNODE) && (speed >= 0.0))
		str += "d";
	else
		str += " ";
	textcol = DefaultObjectColour(trGetElement(signpos.sign), 255);
	TypeLine(dc, textcol, 1, align, newline_at_end, str);
}

// Use: This is a printf-like function that uses a format string to draw a line
// on the device context, and put a newline afterwards.  The format string uses
// forward slashes ('/') to mark commands and glyphs.  The format is:
//
// /g-/    leave a blank space with the size of a glyph
// /gnnn/  print glyph n
// /snnn/  change the scale to n
// /l/ /c/ /r/   change the alignment for the rest of the
//               screen to left, center, or right
// /`/     prints a literal "/"
//
// Example: "/l//s2//gVENUS//s1/ in /s2//gTAURUS/" will draw the glyph of Venus
//          with scale 2, the string " in " in scale 1, and the glyph of Taurus
//          in scale one, all centered vertically and horizontally
void CTypewriterDraw::TypeLine(wxDC &dc, const wxColour &textcol,
							   int initscale, int align,
							   bool newline_at_end, const wxString &fmt,
							   int toth, int totscale)
{
	wxColour col;
	wxString tok, numstr;
	wxStringTokenizer toks;
	wxFont the_font;
	size_t pos;
	bool cmdmode, h_only_mode;
	long n;
	int ww, hh, www, hhh, newh, descent;
	int scale = initscale;
	int bigscale = 1;

	// Use two passes: The first pass to calculate the dimensions, and the second half
	// to actually draw everything

	// Pass #1: Calculating the width and height
	ww = hh = 0;
	bigscale = 1;
	cmdmode = true; // it will be set false at the start of the loop
	h_only_mode = false;
	toks.SetString(fmt, "/");
	while(toks.HasMoreTokens())
	{
		cmdmode = !cmdmode;
		tok = toks.GetNextToken();
		if(tok.length() == 0)
			continue;

		if(cmdmode)
		{
			// Handle a literal "/" as a special case
			if(tok.Cmp("`") == 0)
			{
				the_font = wxFont(glsize*scale, DEFAULT_FAMILY, wxNORMAL,
					              wxNORMAL, FALSE);
				dc.SetFont(the_font);
				dc.GetTextExtent("/", &www, &hhh, NULL, NULL, &the_font);
				if(!h_only_mode) ww += www;
				if(hh < hhh) hh = hhh;
				if(scale > bigscale) bigscale = scale;
				continue;
			}

			// Change the scale if requested
			if(tok.Left(1).CmpNoCase("s") == 0)
			{
				numstr = tok.Right(tok.length() - 1);
				numstr.ToLong(&n);
				if((n > 0) && (n < 1000)) // set some reasonable boundaries!
					scale = n;
				continue;
			}

			// If it's a glyph, factor in its width and height; assume square
			if(tok.Left(1).CmpNoCase("g") == 0)
			{
				if(!h_only_mode) ww += glsize*scale;
				newh = (scale == 2) ? (2*glsize + glsize/2) : 3*glsize*scale/2;
				if(hh < newh) hh = newh;
				if(scale > bigscale) bigscale = scale;
				continue;
			}

			// Stop calculating at a change of alignment; the string
			// will be broken up there anyway
			if((tok.CmpNoCase("l") == 0) || (tok.CmpNoCase("c") == 0) ||
			   (tok.CmpNoCase("r") == 0))
			{
				h_only_mode = true;
				continue;
			}
		}
		else // cmdmode == false
		{
			// factor in the width and height of the string
			the_font = wxFont(glsize*scale, DEFAULT_FAMILY, wxNORMAL,
			                  wxNORMAL, FALSE);
			dc.SetFont(the_font);
			dc.GetTextExtent(tok, &www, &hhh, &descent, NULL, &the_font);
			if(!h_only_mode) ww += www;
			if(hh < (hhh - descent)) hh = hhh - descent;
			if(scale > bigscale) bigscale = scale;
		}
	}
	if(hh < toth) hh = toth;
	if(bigscale < totscale) bigscale = totscale;

	// Pass #2: Do the drawing
	scale = initscale;
	cmdmode = true;
	toks.SetString(fmt, "/");
	// default to centering, as that's what is most common
	if(align == wxNO_ALIGN)
		;
	else if(align == wxLEFT)
		x = x0;
	else if(align == wxRIGHT)
		x = (w - x0 - ww);
	else // center
		x = (w - ww)/2;
	while(toks.HasMoreTokens())
	{
		cmdmode = !cmdmode;
		tok = toks.GetNextToken();
		if(tok.length() == 0)
			continue;
		if(cmdmode)
		{
			// Handle a literal "/" as a special case
			if(tok.Cmp("`") == 0)
			{
				TypeText(dc, "/", textcol, glsize*scale, false,
				         wxLEFT, DEFAULT_FAMILY);
				continue;
			}

			// Change the scale if requested
			if(tok.Left(1).CmpNoCase("s") == 0)
			{
				numstr = tok.Right(tok.length() - 1);
				numstr.ToLong(&n);
				if((n > 0) && (n < 1000)) // set some reasonable boundaries!
					scale = n;
				continue;
			}

			// Draw a glyph
			if(tok.Left(1).CmpNoCase("g") == 0)
			{
				numstr = tok.Right(tok.length() - 1);
				if(numstr.Cmp("-") == 0)
					n = NO_GLYPH;
				else
					numstr.ToLong(&n);
				if(IS_ASPECT(n) || (n == DISPOSITOR_ASPECT))
					col = DefaultAspectColour(n, 255);
				else if((IS_PLANETARY_BODY(n)) || (IS_HOUSE(n)) || (IS_SIGN(n)))
					col = DefaultObjectColour(trGetElement(n), 255);
				else
					col = textcol;
				if(scale != bigscale)
					y += (int)(floor(3.0*(hh - scale*glsize)/4.0 + 0.5));
				TypeGlyph(dc, n, col, scale, (scale <= 2) ? 1 : 2);
				if(scale != bigscale)
					y -= (int)(floor(3.0*(hh - scale*glsize)/4.0 + 0.5));
				continue;
			}

			// Recurse on a change of alignment (note: no newlines, even on "l"!)
			if(tok.CmpNoCase("l") == 0)
			{
				if((pos = fmt.Find('l')) < 0)
					pos = fmt.Find('L');
				if(pos >= 0)
				{
					TypeLine(dc, textcol, scale, wxLEFT,
						newline_at_end, fmt.Right(fmt.length() - pos - 2), hh, bigscale);
					return;
				}
			}
			else if(tok.CmpNoCase("c") == 0)
			{
				if((pos = fmt.Find('c')) < 0)
					pos = fmt.Find('C');
				if(pos >= 0)
				{
					TypeLine(dc, textcol, scale, wxCENTER,
						newline_at_end, fmt.Right(fmt.length() - pos - 2), hh, bigscale);
					return;
				}
			}
			else if(tok.CmpNoCase("r") == 0)
			{
				if((pos = fmt.Find('r')) < 0)
					pos = fmt.Find('R');
				if(pos >= 0)
				{
					TypeLine(dc, textcol, scale, wxRIGHT,
						newline_at_end, fmt.Right(fmt.length() - pos - 2), hh, bigscale);
					return;
				}
			}
		}
		else // cmdmode == false
		{
			the_font = wxFont(glsize*scale, DEFAULT_FAMILY, wxNORMAL,
			                  wxNORMAL, FALSE);
			dc.SetFont(the_font);
			dc.GetTextExtent(tok, &www, &hhh, &descent, NULL, &the_font);

			if(scale != bigscale)
				y += (int)(floor(3.0*(hh - (hhh - descent))/4.0 + 0.5));
			TypeText(dc, tok, textcol, glsize*scale, false, wxLEFT, DEFAULT_FAMILY);
			if(scale != bigscale)
				y -= (int)(floor(3.0*(hh - (hhh - descent))/4.0 + 0.5));
		}
	}
	if(newline_at_end)
		NewLine(hh);
}
